import threading 
import time
import os
import sys
from . import part, multipart_collector, cookie, session
from . import wsgi_executor, xmlrpc_executor, grpc_executor
from skitai.lib import producers
from skitai.server import utility
from hashlib import md5
import random
import base64
from skitai.saddle import cookie
JINJA2 = True
try:	
	from jinja2 import Environment, PackageLoader
except ImportError:
	JINJA2 = False
		
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib


class Config:
	max_post_body_size = 5 * 1024 * 1024
	max_cache_size = 5 * 1024 * 1024
	max_multipart_body_size = 20 * 1024 * 1024
	max_upload_file_size = 20000000


class AuthorizedUser:
	def __init__ (self, user, realm, info = None):
		self.name = user
		self.realm = realm
		self.info = info
		
				
class Saddle (part.Part):
	use_reloader = False
	debug = False
	
	# Session
	securekey = None
	session_timeout = None
	
	#WWW-Authenticate
	authenticate = False
	authorization = "digest"	
	realm = None
	users = {}	
	opaque = None	
	
	def __init__ (self, app_name):
		part.Part.__init__ (self)
		self.app_name = app_name
		self.jinja_env = JINJA2 and Environment (loader = PackageLoader (app_name)) or None		
		self.lock = threading.RLock ()
		self.cache_sorted = 0
		self.cached_paths = {}
		self.cached_rules = []
		self.config = Config ()
	
	def jinja_overlay (self, line_statement = "%", variable_string = "#", block_start_string = "{%", block_end_string = "}", **karg):
		from . import jinjapatch
				
		self.jinja_env = jinjapatch.Environment (
			loader = PackageLoader (self.app_name),
		  variable_start_string=variable_string,
		  variable_end_string=variable_string,
		  line_statement_prefix=line_statement,
		  block_end_string = block_end_string,
		  block_start_string = block_start_string,
		  line_comment_prefix=line_statement * 2,
		  trim_blocks = True,
			lstrip_blocks = True,
		  **karg
		)
	
	def render (self, was, template_file, _do_not_use_this_variable_name_ = {}, **karg):
		while template_file and template_file [0] == "/":
			template_file = template_file [1:]	
											
		if _do_not_use_this_variable_name_: 
			assert not karg, "Can't Use Dictionary and Keyword Args Both"
			karg = _do_not_use_this_variable_name_

		karg ["was"] = was		
		template = self.get_template (template_file)
		self.when_got_template (was, template, karg)
			
		rendered = template.render (karg)
		self.when_template_rendered (was, template, karg, rendered)
		return rendered	
					
	def get_template (self, name):
		if JINJA2:
			return self.jinja_env.get_template (name)
		raise ImportError ("jinja2 required.")
			
	def get_www_authenticate (self):
		if self.authorization == "basic":
			return 'Basic realm="%s"' % self.realm
		else:	
			if self.opaque is None:
				self.opaque = md5 (self.realm.encode ("utf8")).hexdigest ()
			return 'Digest realm="%s", qop="auth", nonce="%s", opaque="%s"' % (
				self.realm, utility.md5uniqid (), self.opaque
			)
	
	def get_password (self, user):
		info = self.users.get (user)
		if not info:
			return None, 0 # passwrod, encrypted
		return type (info) is str and (info, 0) or info [:2]
	
	def get_info (self, user):
		info = self.users.get (user)
		if not info: return None
		return (type (info) is not str and len (info) == 3) and info [-1] or None
				
	def authorize (self, auth, method, uri):
		if self.realm is None or not self.users:
			return
				
		if auth is None:
			return self.get_www_authenticate ()
		
		# check validate: https://evertpot.com/223/
		amethod, authinfo = auth.split (" ", 1)
		if amethod.lower () != self.authorization:
			return self.get_www_authenticate ()
		
		if self.authorization == "basic":
			basic = base64.decodestring (authinfo.encode ("utf8")).decode ("utf8")
			current_user, current_password = basic.split (":", 1)
			password, encrypted = self.get_password (current_user)
			if encrypted:
				raise AssertionError ("Basic authorization can't handle encrypted password")
			if password ==  current_password:
				return AuthorizedUser (current_user, self.realm, self.get_info (current_user))
				
		else:
			method = method.upper ()
			infod = {}
			for info in authinfo.split (","):
				k, v = info.strip ().split ("=", 1)
				if not v: return self.get_www_authenticate ()
				if v[0] == '"': v = v [1:-1]
				infod [k]	 = v
			
			current_user = infod.get ("username")
			if not current_user:
				return self.get_www_authenticate ()
			
			password, encrypted = self.get_password (current_user)
			if not password:
				return self.get_www_authenticate ()
				
			try:
				if uri != infod ["uri"]:					
					return self.get_www_authenticate ()
				if encrypted:	
					A1 = password
				else:
					A1 = md5 (("%s:%s:%s" % (infod ["username"], self.realm, password)).encode ("utf8")).hexdigest ()
				A2 = md5 (("%s:%s" % (method, infod ["uri"])).encode ("utf8")).hexdigest ()
				Hash = md5 (("%s:%s:%s:%s:%s:%s" % (
					A1, 
					infod ["nonce"],
					infod ["nc"],
					infod ["cnonce"],
					infod ["qop"],
					A2
					)).encode ("utf8")
				).hexdigest ()

				if Hash == infod ["response"]:
					return AuthorizedUser (current_user, self.realm, self.get_info (current_user))
					
			except KeyError:
				pass
		
		return self.get_www_authenticate ()
			
	def set_devel (self, debug = True, use_reloader = True):
		self.debug = debug
		self.use_reloader = use_reloader
	
	def get_multipart_collector (self):
		return multipart_collector.MultipartCollector
	
	def get_method (self, path_info, command = None, content_type = None, authorization = None):		
		current_app, method = self, None
		
		with self.lock:
			try:
				method, options = self.cached_paths [path_info]
			except KeyError:
				ind = 0
				for rulepack, freg, options in self.cached_rules:
					method, kargs = self.try_rule (path_info, rulepack)
					if method:						
						self.cached_rules [ind][1] += 1
						break
					ind += 1					
								
		if not method:
			if self.use_reloader:
				self.lock.acquire ()																
			try:	
				current_app, method, kargs, options, match, matchtype = self.get_package_method (path_info, command, content_type, authorization, self.use_reloader)
			finally:	
				if self.use_reloader: 
					self.lock.release ()
			
			if matchtype == -1:
				return current_app, method, None, 301
			
			if not method:
				return current_app, None, None, 404
				
			if not self.use_reloader:
				if matchtype == 1:
					with self.lock:
						self.cached_paths [match] = (method, options)
								
				elif matchtype == 2:
					with self.lock:
						self.cached_rules.append ([match, 1, options])
						if time.time () - self.cache_sorted > 300:
							self.cached_rules.sort (key = lambda x: x[1], reverse = True)
							self.cache_sorted = time.time ()
		
		resp_code = 0
		if options:
			allowed = options.get ("methods", [])
			if allowed and command not in allowed:
				return current_app, None, None, 405 # method not allowed		
			allowed = options.get ("content_types", [])	
			if allowed and content_type not in allowed:
				return current_app, None, None, 415 # unsupported media type
			resp_code = options.get ("authenticate", False) and 401 or 0
									
		return current_app, method, kargs, resp_code			
	
	def restart (self, wasc, route):
		self.start (wasc, route, self.packages)
	
	def create_on_demand (self, was, name):
		class G: 
			pass
		
		# create just in time objects
		if name == "cookie":
			return cookie.Cookie (was.request, self.securekey, self.route [:-1], self.session_timeout)
			
		elif name in ("session", "mbox"):
			if not was.in__dict__ ("cookie"):
				was.cookie = cookie.Cookie (was.request, self.securekey, self.route [:-1], self.session_timeout)			
			if name == "session":
				return was.cookie.get_session ()
			if name == "mbox":
				return was.cookie.get_notices ()
				
		elif name == "g":
			return G ()
		
	def cleanup_on_demands (self, was):
		if was.in__dict__ ("g"):
			del was.g
		if not was.in__dict__ ("cookie"):
			return
		for j in ("session", "mbox"):
			if was.in__dict__ (j):		
				delattr (was, j)
		del was.cookie
										
	def __call__ (self, env, start_response):
		was = env ["skitai.was"]		
		was.app = self
		was.ab = self.build_url
		was.response = was.request.response
		
		content_type = env.get ("CONTENT_TYPE", "")				
		if content_type.startswith ("text/xml") or content_type.startswith ("application/xml"):
			result = xmlrpc_executor.Executor (env, self.get_method) ()
		elif content_type.startswith ("application/grpc"):
			result = grpc_executor.Executor (env, self.get_method) ()
		else:	
			result = wsgi_executor.Executor (env, self.get_method) ()
		
		del was.response
		del was.ab		
		self.cleanup_on_demands (was) # del session, mbox, cookie, g
			
		return result
		


# 2014. 12. 9 by Hans Roh hansroh@gmail.com

VERSION = "0.22.7"
version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  VERSION.split (".")))
NAME = "SWAE/%s.%s" % version_info [:2]

import threading
import sys
HTTP2 = True
try: import h2
except ImportError: HTTP2 = False

WEBSOCKET_REQDATA = 1
WEBSOCKET_DEDICATE = 2
WEBSOCKET_MULTICAST = 3
WEBSOCKET_DEDICATE_THREADSAFE = 4

DB_PGSQL = "*postgresql"
DB_SQLITE3 = "*sqlite3"
DB_REDIS = "*redis"
DB_MONGODB = "*mongodb"

class _WASPool:
	def __init__ (self):
		self.__wasc = None
		self.__p = {}
		
	def __get_id (self):
		return id (threading.currentThread ())
	
	def __repr__ (self):
		return "<class skitai.WASPool at %x>" % id (self)
			
	def __getattr__ (self, attr):
		_was = self._get ()
		if not _was.in__dict__ ("app"):
			# it will be called WSGI middlewares except Saddle,
			# So request object not need
			del _was.request			
		return  getattr (_was, attr)
			
	def __setattr__ (self, attr, value):
		if attr.startswith ("_WASPool__"):
			self.__dict__[attr] = value
		else:	
			setattr (self.__wasc, attr, value)
			for _id in self.__p:
				setattr (self.__p [_id], attr, value)
	
	def __delattr__ (self, attr):
		delattr (self.__wasc, attr)
		for _id in self.__p:
			delattr (self.__p [_id], attr, value)
	
	def _start (self, wasc):
		self.__wasc = wasc
	
	def _del (self):
		_id = self.__get_id ()
		try:
			del self.__p [_id]
		except KeyError:
			pass
				
	def _get (self):
		_id = self.__get_id ()
		try:
			return self.__p [_id]
		except KeyError:
			_was = self.__wasc ()
			self.__p [_id] = _was
			return _was


was = _WASPool ()
def start_was (wasc):
	global was
	was._start (wasc)	

	
def run (**conf):
	from . import lifetime
	from .server import Skitai
	
	class TestServer (Skitai.Loader):
		def __init__ (self, conf):
			self.conf = conf
			Skitai.Loader.__init__ (self, 'test.conf')
			
		def configure (self):
			conf = self.conf
			
			self.set_num_worker (1)
			if conf.get ("certfile"):		
				self.config_certification (conf.get ("certfile"), conf.get ("keyfile"), conf.get ("passphrase"))
			self.config_rcache (100)
			self.config_webserver (
				conf.get ('port', 5000), conf.get ('address', '127.0.0.1'), 
				"Skitai Test Server", conf.get ("certfile") is not None, 
				5, 10
			)
			self.config_threads (conf.get ('threads', 4))
			
			for name, (ctype, members, ssl) in conf.get ("clusters", {}):
				self.add_cluster (ctype, name, members, ssl)
			
			self.install_handler (
				conf.get ("mount"), 
				conf.get ("proxy", False),
				conf.get ("static_max_age", 300)
			)
			lifetime.init ()
	
	if not conf.get ('mount'):
		raise ValueError ('Dictionary mount {mount point: path or app} required')
	
	server = TestServer (conf)
	# timeout for fast keyboard interrupt on win32	
	server.run (2.0)
	
	


from skitai.server  import utility
from . import cookie
from skitai.lib.reraise import reraise 
import sys
from skitai.server.threads import trigger
from skitai.lib.attrdict import AttrDict

def traceback ():	
	t, v, tb = sys.exc_info ()
	tbinfo = []
	assert tb # Must have a traceback
	while tb:
		tbinfo.append((
			tb.tb_frame.f_code.co_filename,
			tb.tb_frame.f_code.co_name,
			str(tb.tb_lineno)
			))
		tb = tb.tb_next

	del tb
	file, function, line = tbinfo [-1]
	return (
		"%s %s, file %s at line %s, %s" % (
			t, v, file, line, 
			function == "?" and "__main__" or "function " + function
		)
	)

		
class Executor:
	def __init__ (self, env, get_method):
		self.env = env	
		self.get_method = get_method
		self.was = None
				
	def chained_exec (self, method, args, karg):
		# recursive before, after, teardown
		# [b, [b, [b, func, s, f, t], s, f, t], s, f, t]
		
		response = None
		[before, func, success, failed, teardown] = method	
		try:
			if before:
				response = before (self.was)
				if response:
					return response
					
			if type (func) is list:
				response = self.chained_exec (func, args, karg)
					
			else:
				response = func (self.was, *args, **karg)
				if type (response) is not list:
					response = [response]											
				
		except MemoryError:
			raise
																										
		except Exception as expt:			
			if failed:
				self.was.traceback ()
				response = failed (self.was, sys.exc_info ())
			if response is None:				
				raise
							
		else:
			success and success (self.was)

		teardown and teardown (self.was)
		return response
		
	def generate_content (self, method, _args, karg):		
		_karg = self.parse_kargs (karg)		
		self.was.request.args = _karg
		response = self.chained_exec (method, _args, _karg)		
		return response
	
	def parse_kargs (self, kargs):
		allkarg = AttrDict ()
		query = self.env.get ("QUERY_STRING")
		data = None
		_input = self.env ["wsgi.input"]
		if _input:
			if type (_input) is dict: # multipart
				self.merge_args (allkarg, _input)
			else:
				data = _input.read ()
		
		if kargs:
			self.merge_args (allkarg, kargs)
		if query: 
			self.merge_args (allkarg, utility.crack_query (query))
		if data:
			self.merge_args (allkarg, utility.crack_query (data))
			
		return allkarg
		
	def merge_args (self, s, n):
		for k, v in list(n.items ()):
			if k in s:
				if type (s [k]) is not list:
					s [k] = [s [k]]
				s [k].append (v)
			else:	 
				s [k] = v
	
	def build_was (self):
		class G:
			pass
			
		_was = self.env["skitai.was"]
		_was.env = self.env
				
		# these objects will be create just in time from _was.
		#_was.cookie = cookie.Cookie (_was.request, _was.app.securekey, _was.app.session_timeout)
		#_was.session = _was.cookie.get_session ()
		#_was.mbox = _was.cookie.get_notices ()
		#_was.g = G ()
		self.was = _was
	
	def commit (self):
		# keep commit order, session -> mbox -> cookie
		if not self.was.in__dict__ ("cookie"):		
			return			
		if self.was.in__dict__ ("session"):		
			self.was.session and self.was.session.commit ()
		if self.was.in__dict__ ("mbox"):
			self.was.mbox and self.was.mbox.commit ()
		self.was.cookie.commit ()
	
	def rollback (self):
		if not self.was.in__dict__ ("cookie"):		
			return			
		# keep commit order, session -> mbox -> cookie
		if self.was.in__dict__ ("session"):		
			self.was.session and self.was.session.rollback ()
		if self.was.in__dict__ ("mbox"):		
			self.was.mbox and self.was.mbox.rollback ()
		self.was.cookie.rollback ()
	
	responses = {
		404: "Not Found",
		415: "Unsupported Content Type",
		405: "Method Not Allowed"			
	}	
	
	def isauthorized (self, app, request):			
		try: 
			www_authenticate = app.authorize (request.get_header ("Authorization"), request.command, request.uri)
			if type (www_authenticate) is str:
				request.response ['WWW-Authenticate'] = www_authenticate
				request.response.send_error ("401 Unauthorized")
				return False				
			elif www_authenticate:
				request.user = www_authenticate
					
		except AttributeError: 
			pass
			
		return True
	
	def find_method (self, request, path):
		current_app, thing, param, respcode = self.get_method (
			path, 
			request.command.upper (), 
			request.get_header ('content-type'),
			request.get_header ('authorization')
		)
		
		if respcode == 401 and self.isauthorized (current_app, request):
			# passed then be normal
			respcode = 0
		
		if respcode:
			if respcode == 301:
				request.response ["Location"] = thing
				request.response.send_error ("301 Object Moved", why = 'Object Moved To <a href="%s">Here</a>' % thing)							
			else:
				request.response.send_error ("%d %s" % (respcode, self.responses.get (respcode, "Undefined Error")))
			
		return current_app, thing, param, respcode
		
	def __call__ (self):		
		request = self.env ["skitai.was"].request
		current_app, thing, param, respcode = self.find_method (request, self.env ["PATH_INFO"])
		
		if respcode: 
			# unacceptable
			return b""
		
		self.build_was ()
		self.was.subapp = current_app
		try:
			content = self.generate_content (thing, (), param)
		except:				
			self.rollback ()
			content = request.response ("500 Internal Server Error", exc_info = self.was.app.debug and sys.exc_info () or None)
			raise
		else:
			self.commit ()
		
		# clean was
		del self.was.env
		del self.was.subapp
		return content
		


import re
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
from skitai.server import http_date
from skitai.lib import compressors
import time
import json
import struct
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO

class RepsonseError (Exception): 
	pass

RESPONSE = re.compile ('HTTP/([0-9.]+) ([0-9]{3})\s?(.*)')
def crack_response (data):
	global RESPONSE
	[ version, code, msg ] = RESPONSE.findall(data)[0]
	return version, int(code), msg


class FakeParser(object):
	def __init__(self, target):
		self.target = target

	def feed(self, data):
		self.target.feed(data)

	def close(self):
		pass


class cache_wrap_buffer:
	def __init__ (self, buf, cache):
		self.buf = buf
		self.cache = cache
		self.cdata = None
		
	def __getattr__ (self, name):
		return getattr (self.buf, name)
		
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.buf.close ()
		if self.cache:
			self.cdata = res
		return res
	
	def no_cache (self):
		self.cache = 0
		self.cdata = []
			
	
class list_buffer (cache_wrap_buffer):
	def __init__(self, cache = 0):
		self.cache = cache
		self.data = []
		self.cdata = None

	def feed(self, data):
		self.data.append(data)
	
	def build_data (self):
		return b''.join(self.data)
	
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.build_data ()
		if self.cache:
			self.cdata = res
		return res
	
	def no_cache (self):
		self.cache = 0
		self.cdata = []


class bytes_buffer:
	def __init__(self, cache = 0):
		self.cache = cache
		self.fp = BytesIO ()
		self.cdata = None
		
	def feed (self, data):
		self.fp.write (data)
	
	def get_messages (self):
		msgs = []
		fp = self.fp
		
		fp.seek (0)
		byte = fp.read (1)
		while byte:
			iscompressed = struct.unpack ("<B", byte) [0]
			length = struct.unpack ("<i", fp.read (4)) [0]
			msg = fp.read (length)
			msgs.append (msg)
			byte = fp.read (1)
		return tuple (msgs)
		
	def build_data (self):
		return self.fp.getvalue ()
	
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.build_data ()
		if self.cache:
			self.cdata = res
		return res
	
	def no_cache (self):
		self.cache = 0
		self.cdata = []


class grpc_buffer (bytes_buffer):
	def __init__(self, cache = 0):
		self.cache = cache
		self.fp = BytesIO ()
		self.cdata = None
		
	def feed (self, data):
		self.fp.write (data)
	
	def get_messages (self):
		msgs = []
		fp = self.fp
		
		fp.seek (0)
		byte = fp.read (1)
		while byte:
			iscompressed = struct.unpack ("<B", byte) [0]
			length = struct.unpack ("<i", fp.read (4)) [0]
			msg = fp.read (length)
			msgs.append (msg)
			byte = fp.read (1)
		return tuple (msgs)
		
	def build_data (self):
		return self.get_messages ()
	
	def close (self):
		if self.cdata:
			return self.cdata
		res = self.build_data ()
		if self.cache:
			self.cdata = res
		return res
	
	def no_cache (self):
		self.cache = 0
		self.cdata = []
		
				

def getfakeparser (target_class, cache = False):
	target = target_class(cache)
	return FakeParser(target), target


class Response:
	SIZE_LIMIT = 2**19
	
	def __init__ (self, request, header):		
		self.request = request		
		if header [:2] == "\r\n":
			header = header [2:]
		header = header.split ("\r\n")	
		self.response = header [0]
		self.header = header [1:]
		self._header_cache = {}
		self.version, self.code, self.msg = crack_response (self.response)
		self.size = 0
		self.got_all_data = False
		self.max_age = 0
		self.decompressor = None
		self.is_xmlrpc_return = False
		
	def set_max_age (self):
		self.max_age = 0
		if self.code != 200:
			return
		
		expires = self.get_header ("expires")		
		if expires:
			try:
				val = http_date.parse_http_date (expires)
			except:
				val = 0
	
			if val:
				max_age = val - time.time ()
				if max_age > 0:
					self.max_age = int (max_age)
					return
	
		cache_control = self.get_header ("cache-control")
		if not cache_control:
			return
			
		for each in cache_control.split (","):			
			try: 
				k, v = each.split("=")					
				if k.strip () == "max-age":
					max_age  = int (v)
					if max_age > 0:
						self.max_age = max_age
						break
			except ValueError: 
				continue
		
		if self.max_age > 0:
			age = self.get_header ("age")
			if age:
				try: age = int (age)	
				except: pass	
				else:
					self.max_age -= age
				
	def done (self):
		# it must be called finally
		self.got_all_data = True
		
		if self.decompressor:
			try:
				data = self.decompressor.flush ()
			except:					
				pass
			else:
				self.p.feed (data)			
			self.decompressor = None
	
	def init_buffer (self):
		self.set_max_age ()
		ct = self.get_header ("content-type", "")
		if ct.startswith ('application/grpc'):
			self.p, self.u = getfakeparser (grpc_buffer, cache = self.max_age)
		elif ct == 'text/xml' and self.request.xmlrpc_serialized ():
			self.p, self.u = xmlrpclib.getparser()
			self.u = cache_wrap_buffer (self.u, self.max_age)
			self.is_xmlrpc_return = True
		else:			
			self.p, self.u = getfakeparser (bytes_buffer, cache = self.max_age)
					
		if self.get_header ("Content-Encoding") == "gzip":			
			self.decompressor = compressors.GZipDecompressor ()
			
	def collect_incoming_data (self, data):
		if self.size == 0:
			self.init_buffer ()		
		self.size += len (data)
			
		#print ("<<<<<", repr (data)[-80:], len (data) )	
		if self.decompressor:
			data = self.decompressor.decompress (data)
		
		if self.max_age and self.size > self.SIZE_LIMIT:
			self.max_age = 0
			self.u.no_cache ()
		
		#print (">>>>>", repr (data)[-80:])
		if data:
			# sometimes decompressor return "",
			# null byte is signal of producer's ending, so ignore.
			self.p.feed (data)
	
	def get_header_with_attr (self, header, default = None):
		d = {}
		v = self.get_header (header)
		if v is None:
			return default, d
			
		v2 = v.split (";")
		if len (v2) == 1:
			return v, d
		for each in v2 [1:]:
			try:
				a, b = each.strip ().split ("=", 1)
			except ValueError:
				a, b = each.strip (), None
			d [a] = b
		return v2 [0], d	
				
	def get_header (self, header = None, default = None):
		if header is None:
			return self.header
		header = header.lower()
		hc = self._header_cache
		if header not in hc:
			h = header + ':'
			hl = len(h)
			for line in self.header:
				if line [:hl].lower() == h:
					r = line [hl:].strip ()
					hc [header] = r
					return r
			hc [header] = None
			return default
		else:
			return hc[header] is not None and hc[header] or default
	
	def get_headers (self):
		return self.header
	
	def get_content (self):
		if self.code >= 700:
			return None
		
		if self.size == 0:
			return b""
				
		self.p.close ()
		result = self.u.close()	
		ct = self.get_header ("content-type")	
		if self.is_xmlrpc_return:
			if len(result) == 1:
				result = result [0]
			return result
		elif ct.startswith ("application/json"):
			return json.loads (result)
		return result
	

class FailedResponse (Response):
	def __init__ (self, errcode, msg, request = None):
		self.version, self.code, self.msg, self.header = "1.0", errcode, msg, []
		self.request = request
		self.buffer = None
		self.got_all_data = True
		self.max_age = 0
				
	def collect_incoming_data (self, data):
		raise IOError("This Is Failed Response")
	
	def more (self):
		return b""
		
	def get_content (self):
		return b""
	
	def done (self):
		pass
		


import time
from skitai.server.threads import socket_map
from skitai.server.threads import trigger
import threading
from skitai.server.rpc import cluster_dist_call, rcache
from skitai.lib.attrdict import AttrDict
import asyncore
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB

class OperationTimeout (Exception):
	pass

class Result (rcache.Result):
	def __init__ (self, id, status, description = None, expt_class = None, expt_str = None, data = [], ident = None):
		rcache.Result.__init__ (self, status, ident)		
		self.node = id
		self.__expt_class = expt_class
		self.__expt_str = expt_str
						
		if status == 3:
			try:
				self.set_result (description, data)
			except:
				self.__expt_class, self.__expt_str = asyncore.compact_traceback() [1:3]
				self.status = 2
		
		# For Results competitable		
		if self.status == 3:
			self.code, self.msg = 200, "OK"
		else:
			self.code, self.msg = 501, "Server Error"
		
	def __iter__ (self):
		return self.data.__iter__ ()
	
	def __slice__(self, start = None, end = None, step = None): 
		return self.data [slice(start, end, step)]
		
	def set_result (self, description, data):		
		if not data:
			self.data = data
			return
		
		self.description = description
		if description:
			assert (len (description) == len (data [0]))		
			cols = [type (col) is tuple and col [0] or col.name for col in description]		
			d = []
			for row in data:
				i = 0
				drow = AttrDict ()
				for name in cols:
					drow [name] = row [i]
					i += 1
				d.append (drow)
			self.data = d
			
		else:
			if type (data) is dict:
				try:
					self.data = AttrDict ()
					for k, v in data.items ():
						self.data [k] = v
				except:
					self.data = data
			else:		
				self.data = data
	
	def reraise (self):
		if self.__expt_class:
			raise self.__expt_class ("%s (status: %d)" % (self.__expt_str, self.status))
	
	def get_error_as_string (self):
		if self.__expt_class:
			return "%s %s" % (self.__expt_class, self.__expt_str)		
		
	def cache (self, timeout = 300):
		if self.status != 3:
			return
		rcache.Result.cache (self, timeout)
	
			
class Dispatcher:
	def __init__ (self, cv, id, ident = None, filterfunc = None):
		self.__cv = cv
		self.id = id
		self.ident = ident
		self.filterfunc = filterfunc
		self.creation_time = time.time ()
		self.status = 0
		self.result = None
			
	def get_id (self):
		return self.id
	
	def get_status (self):
		# 0: Not Connected
		# 1: Operation Timeout
		# 2: Exception Occured
		# 3: Normal
		self.__cv.acquire ()		
		status = self.status
		self.__cv.release ()
		return status
		
	def set_status (self, code):
		self.__cv.acquire ()
		self.status = code
		self.__cv.notify ()
		self.__cv.release ()
		return code
		
	def get_result (self):
		if self.result is None:
			self.result = Result (self.id, 1, None, OperationTimeout, "Operation Timeout", None, self.ident)
			self.do_filter ()
		return self.result
	
	def do_filter (self):
		if self.filterfunc:
			self.filterfunc (self.result)
						
	def handle_result (self, description = None, expt_class = None, expt_str = None, data = None):
		if self.get_status () == 1:
			# timeout, ignore
			return
				
		if expt_class:
			status = 2			
		else:
			status = 3
		
		self.result = Result (self.id, status, description, expt_class, expt_str, data, self.ident)
		self.set_status (status)		
		        	     
#-----------------------------------------------------------
# Cluster Base Call
#-----------------------------------------------------------

class ClusterDistCall (cluster_dist_call.ClusterDistCall):
	def __init__ (self, 
		cluster,
		server, 
		dbname, 
		user, 
		password,
		dbtype,
		use_cache,
		mapreduce,
		filter,
		callback,
		logger
		):
		
		self.server = server
		self.dbname = dbname
			
		self.user = user
		self.password = password
		self.dbtype = dbtype
		
		if self.dbtype == DB_PGSQL:
			if self.dbname in (DB_SQLITE3, DB_REDIS):
				self.dbtype = self.dbname
				self.dbname = ""
			elif self.user in (DB_MONGODB,):
				self.dbtype = self.user
				self.user = ""
				
		self._use_cache = use_cache		
		self._filter = filter
		self._callback = callback
		self._mapreduce = mapreduce		
		self._cluster = cluster
		self._cached_request_args = None
		self._cached_result = None
		
		self._logger = logger		
		self._requests = {}
		self._results = []
		self._canceled = 0
		self._init_time = time.time ()
		self._cv = None
		self._retry = 0	
		self._numnodes = 0
		
		self._sent_result = None
		
		if self._cluster:
			nodes = self._cluster.get_nodes ()
			self._numnodes = len (nodes)
			if self._mapreduce:
				self._nodes = nodes
			else: # anyone of nodes
				self._nodes = [None]
		
	def _get_ident (self):
		cluster_name = self._cluster.get_name ()
		if cluster_name == "dbpool":
			_id = "%s/%s/%s/%s" % (self.server, self.dbname, self.user, self.password)
		else:
			_id = cluster_name
		_id += "/%s%s" % (
			",".join (map (lambda x: str (x), self._cached_request_args)),
			self._mapreduce and "/M" or ""
			)	
		return _id
				
	def _get_connection (self, id = None):
		if id is None: id = self._nodes.pop ()
		else: self._nodes = []
			
		if self.server:
			asyncon = self._cluster.get (self.server, self.dbname, self.user, self.password, self.dbtype)		
		else:	
			asyncon = self._cluster.get (id)
		
		if self._cv is None:
			self._cv = asyncon._cv
		return asyncon
	
	def _request (self, method, cmd):
		self._cached_request_args = cmd # backup for retry
		if self._use_cache and rcache.the_rcache:
			self._cached_result = rcache.the_rcache.get (self._get_ident ())
			if self._cached_result is not None:
				return
		
		while self._avails ():
			asyncon = self._get_connection (None)			
			rs = Dispatcher (self._cv, asyncon.address, ident = not self._mapreduce and self._get_ident () or None, filterfunc = self._filter)
			self._requests [rs] = asyncon
			print ('---', method, cmd)
			if method in ("do", "execute"):
				asyncon.execute (self._callback and self._callback or rs.handle_result, *cmd)							
			else:
				asyncon.execute (self._callback and self._callback or rs.handle_result, method, *cmd)			
		trigger.wakeup ()
		return self
	

class ClusterDistCallCreator:
	def __init__ (self, cluster, logger):
		self.cluster = cluster
		self.logger = logger
	
	def __getattr__ (self, name):	
		return getattr (self.cluster, name)
		
	def Server (self, server = None, dbname = None, user = None, password = None, dbtype = None, use_cache = True, mapreduce = False, filter = None, callback = None):
		# reqtype: xmlrpc, rpc2, json, jsonrpc, http
		#return ClusterDistCall (self.cluster, server, dbname, user, password, dbtype, use_cache, mapreduce, filter, callback, self.logger)
		return cluster_dist_call.Proxy (ClusterDistCall, self.cluster, server, dbname, user, password, dbtype, use_cache, mapreduce, filter, callback, self.logger)


		
	
if __name__ == "__main__":
	from skitai.lib  import logger
	from . import cluster_manager
	import sys
	import asyncore
	import time
	from skitai.client import socketpool
	
	def _reduce (asyncall):
		for rs in asyncall.getswait (5):
			print("Result:", rs.id, rs.status, rs.code, repr(rs.result [:60]))
					
	def testCluster ():	
		sc = cluster_manager.ClusterManager ("tt", ["210.116.122.187:3424 1", "210.116.122.184:3424 1", "175.115.53.148:3424 1"], logger= logger.screen_logger ())
		clustercall = ClusterDistCallCreator (sc, logger.screen_logger ())	
		s = clustercall.Server ("rpc2", login = "admin/whddlgkr")
		s.bladese.util.status ("openfos.v2")		
		threading.Thread (target = _reduce, args = (s,)).start ()
		
		while 1:
			asyncore.loop (timeout = 1, count = 2)
			if len (asyncore.socket_map) == 1:
				break
	
	def testSocketPool ():
		sc = socketpool.SocketPool (logger.screen_logger ())
		clustercall = ClusterDistCallCreator (sc, logger.screen_logger ())			
		s = clustercall.Server ("http://www.bidmain.com/")
		s.request ()
		
		#s = clustercall.Server ("http://210.116.122.187:3424/rpc2", "admin/whddlgkr")
		#s.bladese.util.status ("openfos.v2")
		
		threading.Thread (target = __reduce, args = (s,)).start ()
		
		while 1:
			asyncore.loop (timeout = 1, count = 2)
			print(asyncore.socket_map)
			if len (asyncore.socket_map) == 1:
				break
	
	trigger.start_trigger ()
	
	testCluster ()
	testSocketPool ()	



import asynchat

class fake_channel:
	def __init__ (self, channel):
		# override members
		self._channel = channel
		self.addr = channel.addr
		self.connected = channel.connected
	
	def __getattr__ (self, attr):
		return getattr (self._channel, attr)


class data_channel (fake_channel, asynchat.async_chat):
	def __init__ (self, channel, content_length):
		asynchat.async_chat.__init__ (self)
		fake_channel.__init__ (self, channel)
		self._content_length = content_length
		self._data = b""		
		self._data_size = 0
		self._chunks  = []
			
	def set_data (self, data, size):
		self._data = data
		self.current_request.rbytes = len (self._data)
		self._data_size += size
		self._chunks.append (size)	
		self.handle_read ()		
	
	def get_chunk_size (self):
		d = {}
		for c in self._chunks:
			d [c] = None
		return len (d) == 1 and c
			
	def get_data_size (self):
		return self._data_size

	def get_content_length (self):
		return self._content_length
		
	def recv (self, buffer_size):
		data, self._data = self._data, b""
		return data
				
	def collect_incoming_data (self, data):	
		self.current_request.collect_incoming_data (data)
	
	def found_terminator (self):
		self.current_request.found_terminator ()
	


# -*- Mode: Python; tab-width: 4 -*-
#	$Id: filesys.py,v 1.10 2001/04/16 03:45:59 rushing Exp $
#	Author: Sam Rushing <rushing@nightmare.com>
#
# Generic filesystem interface.
#

# We want to provide a complete wrapper around any and all
# filesystem operations.

# this class is really just for documentation,
# identifying the API for a filesystem object.

# opening files for reading, and listing directories, should
# return a producer.

class abstract_filesystem:
	def __init__ (self):
		pass

	def current_directory (self):
		"Return a string representing the current directory."
		pass

	def listdir (self, path, long=0):
		"""Return a listing of the directory at 'path' The empty string
		indicates the current directory.  If 'long' is set, instead
		return a list of (name, stat_info) tuples
		"""
		pass

	def open (self, path, mode):
		"Return an open file object"
		pass

	def stat (self, path):
		"Return the equivalent of os.stat() on the given path."
		pass

	def isdir (self, path):
		"Does the path represent a directory?"
		pass

	def isfile (self, path):
		"Does the path represent a plain file?"
		pass

	def cwd (self, path):
		"Change the working directory."
		pass

	def cdup (self):
		"Change to the parent of the current directory."
		pass


	def longify (self, path):
		"""Return a 'long' representation of the filename
		[for the output of the LIST command]"""
		pass

# standard wrapper around a unix-like filesystem, with a 'false root'
# capability.

# security considerations: can symbolic links be used to 'escape' the
# root?  should we allow it?  if not, then we could scan the
# filesystem on startup, but that would not help if they were added
# later.  We will probably need to check for symlinks in the cwd method.

# what to do if wd is an invalid directory?

import os
import stat

import string

def safe_stat (path):
	try:
		return (path, os.stat (path))
	except:
		return None

import re
import glob

class os_filesystem:
	path_module = os.path

	# set this to zero if you want to disable pathname globbing.
	# [we currently don't glob, anyway]
	do_globbing = 1

	def __init__ (self, root, wd='/'):
		self.root = root
		self.wd = wd

	def current_directory (self):
		return self.wd

	def isfile (self, path):
		p = self.normalize (self.path_module.join (self.wd, path))				
		try:
			return self.path_module.isfile (self.translate(p))
		except TypeError:
			return False	

	def isdir (self, path):
		p = self.normalize (self.path_module.join (self.wd, path))
		try:
			return self.path_module.isdir (self.translate(p))
		except TypeError:
			return False
				
	def cwd (self, path):
		p = self.normalize (self.path_module.join (self.wd, path))
		translated_path = self.translate(p)
		if not self.path_module.isdir (translated_path):
			return 0
		else:
			old_dir = os.getcwd()
			# temporarily change to that directory, in order
			# to see if we have permission to do so.
			try:
				can = 0
				try:
					os.chdir (translated_path)
					can = 1
					self.wd = p
				except:
					pass
			finally:
				if can:
					os.chdir (old_dir)
			return can

	def cdup (self):
		return self.cwd ('..')

	def listdir (self, path, long=0):
		p = self.translate (path)
		# I think we should glob, but limit it to the current
		# directory only.
		ld = os.listdir (p)
		if not int:
			return list_producer (ld, 0, None)
		else:
			old_dir = os.getcwd()
			try:
				os.chdir (p)
				# if os.stat fails we ignore that file.
				result = [_f for _f in map (safe_stat, ld) if _f]
			finally:
				os.chdir (old_dir)
			return list_producer (result, 1, self.longify)

	# TODO: implement a cache w/timeout for stat()
	def stat (self, path):
		p = self.translate (path)
		return os.stat (p)

	def open (self, path, mode):
		p = self.translate (path)		
		return open (p, mode)

	def unlink (self, path):
		p = self.translate (path)
		return os.unlink (p)

	def mkdir (self, path):
		p = self.translate (path)
		return os.mkdir (p)

	def rmdir (self, path):
		p = self.translate (path)
		return os.rmdir (p)

	# utility methods
	def normalize (self, path):
		# watch for the ever-sneaky '/+' path element
		path = re.sub ('/+', '/', path)
		p = self.path_module.normpath (path)
		# remove 'dangling' cdup's.
		if len(p) > 2 and p[:3] == '/..':
			p = '/'
		return p

	def translate (self, path):
		# we need to join together three separate
		# path components, and do it safely.
		# <real_root>/<current_directory>/<path>
		# use the operating system's path separator.
		path = string.join (string.split (path, '/'), os.sep)
		p = self.normalize (self.path_module.join (self.wd, path))
		p = self.normalize (self.path_module.join (self.root, p[1:]))
		return p

	def longify (self, xxx_todo_changeme):
		(path, stat_info) = xxx_todo_changeme
		return unix_longify (path, stat_info)

	def __repr__ (self):
		return '<unix-style fs root:%s wd:%s>' % (
			self.root,
			self.wd
			)


class mapped_filesystem (os_filesystem):
	def __init__ (self):
		os_filesystem.__init__ (self, None, "/")
		self.maps = {}
		self.permission_cache = {}
	
	def add_map (self, alias, optstr):
		options = optstr.split("|")		
		
		path = options [0]
		if alias == "":
			os_filesystem.__init__ (self, path)
		elif alias [-1] == "/":
			alias = alias [:-1]

		conf = {"path": path}
		for opt in options [1:]:
			maybe_pair = opt.split ("=", 1)
			if len (maybe_pair) == 1:
				conf [opt.strip ()] = None
			else:
				optk, optv = maybe_pair[0].strip (), maybe_pair[1].strip ()
				if optv [:5] == "list:":
					conf [optk] = [x.strip () for x in optv [5:].split(",")]
				else:	
					conf [optk] = optv		
		self.maps [alias] = conf
		
		if self.root is None:
			os_filesystem.__init__ (self, None) # init forcely
			
	def translate (self, path):
		if not self.maps: 
			return None
		
		p = self.normalize (self.path_module.join (self.wd, path))
		path = p.split (os.sep)
		
		maybe_alias = ""
		current_depth = 0
		latest_alias = ""
		latest_depth = ""
		
		for each in path [1:]:
			current_depth += 1
			maybe_alias += "/" + each			
			
			if maybe_alias in self.maps:
				latest_alias = maybe_alias
				latest_depth = current_depth
								
		if latest_alias:			
			return self.normalize (self.path_module.join (self.maps [latest_alias]["path"], '/'.join (path [latest_depth + 1:])))
		if self.root is None:
			return None # no psysical root
		return self.normalize (self.path_module.join (self.root, p[1:]))
			
	def get_permission (self, path):
		try:
			return self.permission_cache [path]
		except KeyError:
			pass
				
		p = self.normalize (self.path_module.join (self.wd, path))
		patheach = p.split (os.sep)				
		maybe_alias = ""
		collected_permissions = []
		
		for each in patheach [1:]:
			maybe_alias += "/" + each			
			if maybe_alias in self.maps:
				collected_permissions.append (self.maps [maybe_alias].get ("permission", []))
		
		collected_permissions = [_f for _f in collected_permissions if _f]		
		if not collected_permissions:
			permission = []
		else:
			permission = collected_permissions [-1]
		
		self.permission_cache [path] = permission
		return permission
	
			
if os.name == 'posix':

	class unix_filesystem (os_filesystem):
		pass

	class schizophrenic_unix_filesystem (os_filesystem):
		PROCESS_UID		= os.getuid()
		PROCESS_EUID	= os.geteuid()
		PROCESS_GID		= os.getgid()
		PROCESS_EGID	= os.getegid()

		def __init__ (self, root, wd='/', persona=(None, None)):
			os_filesystem.__init__ (self, root, wd)
			self.persona = persona

		def become_persona (self):
			if self.persona is not (None, None):
				uid, gid = self.persona
				# the order of these is important!
				os.setegid (gid)
				os.seteuid (uid)

		def become_nobody (self):
			if self.persona is not (None, None):
				os.seteuid (self.PROCESS_UID)
				os.setegid (self.PROCESS_GID)

		# cwd, cdup, open, listdir
		def cwd (self, path):
			try:
				self.become_persona()
				return os_filesystem.cwd (self, path)
			finally:
				self.become_nobody()

		def cdup (self, path):
			try:
				self.become_persona()
				return os_filesystem.cdup (self)
			finally:
				self.become_nobody()

		def open (self, filename, mode):
			try:
				self.become_persona()
				return os_filesystem.open (self, filename, mode)
			finally:
				self.become_nobody()

		def listdir (self, path, long=0):
			try:
				self.become_persona()
				return os_filesystem.listdir (self, path, int)
			finally:
				self.become_nobody()

# This hasn't been very reliable across different platforms.
# maybe think about a separate 'directory server'.
#
#	import posixpath
#	import fcntl
#	import FCNTL
#	import select
#	import asyncore
#
#	# pipes /bin/ls for directory listings.
#	class unix_filesystem (os_filesystem):
#		pass
# 		path_module = posixpath
#
# 		def listdir (self, path, long=0):
# 			p = self.translate (path)
# 			if not long:
# 				return list_producer (os.listdir (p), 0, None)
# 			else:
# 				command = '/bin/ls -l %s' % p
# 				print 'opening pipe to "%s"' % command
# 				fd = os.popen (command, 'rt')
# 				return pipe_channel (fd)
#
# 	# this is both a dispatcher, _and_ a producer
# 	class pipe_channel (asyncore.file_dispatcher):
# 		buffer_size = 4096
#
# 		def __init__ (self, fd):
# 			asyncore.file_dispatcher.__init__ (self, fd)
# 			self.fd = fd
# 			self.done = 0
# 			self.data = ''
#
# 		def handle_read (self):
# 			if len (self.data) < self.buffer_size:
# 				self.data = self.data + self.fd.read (self.buffer_size)
# 			#print '%s.handle_read() => len(self.data) == %d' % (self, len(self.data))
#
# 		def handle_expt (self):
# 			#print '%s.handle_expt()' % self
# 			self.done = 1
#
# 		def ready (self):
# 			#print '%s.ready() => %d' % (self, len(self.data))
# 			return ((len (self.data) > 0) or self.done)
#
# 		def more (self):
# 			if self.data:
# 				r = self.data
# 				self.data = ''
# 			elif self.done:
# 				self.close()
# 				self.downstream.finished()
# 				r = ''
# 			else:
# 				r = None
# 			#print '%s.more() => %s' % (self, (r and len(r)))
# 			return r

# For the 'real' root, we could obtain a list of drives, and then
# use that.  Doesn't win32 provide such a 'real' filesystem?
# [yes, I think something like this "\\.\c\windows"]

class msdos_filesystem (os_filesystem):
	def longify (self, xxx_todo_changeme1):
		(path, stat_info) = xxx_todo_changeme1
		return msdos_longify (path, stat_info)

# A merged filesystem will let you plug other filesystems together.
# We really need the equivalent of a 'mount' capability - this seems
# to be the most general idea.  So you'd use a 'mount' method to place
# another filesystem somewhere in the hierarchy.

# Note: this is most likely how I will handle ~user directories
# with the http server.

class merged_filesystem:
	def __init__ (self, *fsys):
		pass

# this matches the output of NT's ftp server (when in
# MSDOS mode) exactly.

def msdos_longify (file, stat_info):
	if stat.S_ISDIR (stat_info[stat.ST_MODE]):
		dir = '<DIR>'
	else:
		dir = '     '
	date = msdos_date (stat_info[stat.ST_MTIME])
	return '%s       %s %8d %s' % (
		date,
		dir,
		stat_info[stat.ST_SIZE],
		file
		)

def msdos_date (t):
	try:
		info = time.gmtime (t)
	except:
		info = time.gmtime (0)
	# year, month, day, hour, minute, second, ...
	if info[3] > 11:
		merid = 'PM'
		info[3] = info[3] - 12
	else:
		merid = 'AM'
	return '%02d-%02d-%02d  %02d:%02d%s' % (
		info[1],
		info[2],
		info[0]%100,
		info[3],
		info[4],
		merid
		)

months = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
		  'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

mode_table = {
	'0':'---',
	'1':'--x',
	'2':'-w-',
	'3':'-wx',
	'4':'r--',
	'5':'r-x',
	'6':'rw-',
	'7':'rwx'
	}

import time

def unix_longify (file, stat_info):
	# for now, only pay attention to the lower bits
	mode = ('%o' % stat_info[stat.ST_MODE])[-3:]
	mode = string.join ([mode_table[x] for x in mode], '')
	if stat.S_ISDIR (stat_info[stat.ST_MODE]):
		dirchar = 'd'
	else:
		dirchar = '-'
	date = ls_date (int(time.time()), stat_info[stat.ST_MTIME])
	return '%s%s %3d %-8d %-8d %8d %s %s' % (
		dirchar,
		mode,
		stat_info[stat.ST_NLINK],
		stat_info[stat.ST_UID],
		stat_info[stat.ST_GID],
		stat_info[stat.ST_SIZE],
		date,
		file
		)
		
# Emulate the unix 'ls' command's date field.
# it has two formats - if the date is more than 180
# days in the past, then it's like this:
# Oct 19  1995
# otherwise, it looks like this:
# Oct 19 17:33

def ls_date (now, t):
	try:
		info = time.gmtime (t)
	except:
		info = time.gmtime (0)
	# 15,600,000 == 86,400 * 180
	if (now - t) > 15600000:
		return '%s %2d  %d' % (
			months[info[1]-1],
			info[2],
			info[0]
			)
	else:
		return '%s %2d %02d:%02d' % (
			months[info[1]-1],
			info[2],
			info[3],
			info[4]
			)

# ===========================================================================
# Producers
# ===========================================================================

class list_producer:
	def __init__ (self, file_list, long, longify):
		self.file_list = file_list
		self.long = int
		self.longify = longify
		self.done = 0

	def ready (self):
		if len(self.file_list):
			return 1
		else:
			if not self.done:
				self.done = 1
			return 0
		return (len(self.file_list) > 0)

	# this should do a pushd/popd
	def more (self):
		if not self.file_list:
			return ''
		else:
			# do a few at a time
			bunch = self.file_list[:50]
			if self.long:
				bunch = list(map (self.longify, bunch))
			self.file_list = self.file_list[50:]
			return string.joinfields (bunch, '\r\n') + '\r\n'




import threading

class priority_producer_fifo:
	def __init__ (self):
		self.l = []
		self._lock = threading.Lock ()
	
	def __len__ (self):
		with self._lock:
			l = len (self.l)
		return l
		
	def __getitem__(self, index):
		with self._lock:
			i = self.l [index]
		return i	
	
	def __setitem__(self, index, item):
		with self._lock:
			self.l.insert (index, item)
		
	def __delitem__ (self, index):	
		with self._lock:
			del self.l [index]
	
	def remove (self, stream_id):	
		index = 0
		with self._lock:
			for each in self.l:
				try:
					legacy_stream_id = each.stream_id
				except AttributeError:
					pass
				else:
					if legacy_stream_id	== stream_id:
						item = self.l.pop (index)
						if hasattr (item, 'close'):
							try: item.close ()
							except: pass
						break
				index += 1		
			
	def append (self, item):
		has_None = False
		with self._lock:
			if self.l and self.l [-1] is None:
				has_None = True
		if has_None:
			return
				
		if item is None:
			with self._lock:
				self.l.append (item)
			return
			
		try:
			w1 = item.weight
			d1 = item.depends_on
		except AttributeError:
			with self._lock:
				self.l.append (item)
			return
		
		index = 0
		inserted = False
		with self._lock:
			for each in self.l:
				try:
					w2 = each.weight
					d2 = each.depends_on
				except AttributeError:
					pass
				else:
					if d2 >= d1 and w2 < w1:
						self.l.insert (index, item)
						inserted = True
						break
				index += 1
		
		if not inserted:
			with self._lock:
				self.l.append (item)
			
	def appendleft (self, item):
		with self._lock:
			self.l.insert (0, item)
		
	def clear (self):
		with self._lock:
			self.l = []
	
	def ready (self):
		# check if remote flow control window for each stream is open
		# but maybe unnecessory and unused
		self._lock.acquire ()
		if not self.l:
			self._lock.release ()
			return False
			
		first = self.l [0]
		if not hasattr (first, 'stream_id') or first.ready ():
			self._lock.release ()
			return True
		
		index = 1
		has_ready = False
		for item in self.l [1:]:
			if not hasattr (item, 'stream_id'):
				break
							
			if item.ready ():
				ready_item = self.l.pop (index)
				self.l.insert (0, ready_item)
				has_ready = True
				break
				
			index += 1
		
		self._lock.release ()
		return has_ready
				


from skitai.protocol.http import response

class Response (response.Response):
	def __init__ (self, request, code, msg, opcode, data = None):
		self.request = request
		self.code = code
		self.msg = msg
		self.data = data
		self.version = "1.1"
		self.header = ["OPCODE: %s" % opcode]
	
	def get_content (self):
		return self.data
	
	def done (self):
		pass
	


from . import wsgi_handler
import skitai
from skitai.lib import producers
from h2.connection import H2Connection, GoAwayFrame, DataFrame
from h2.exceptions import ProtocolError, NoSuchStreamError
from h2.events import DataReceived, RequestReceived, StreamEnded, PriorityUpdated, ConnectionTerminated, StreamReset, WindowUpdated
from h2.errors import PROTOCOL_ERROR, FLOW_CONTROL_ERROR, NO_ERROR
from .http2.request import request as http2_request
from .http2.vchannel import fake_channel, data_channel
from skitai.protocol.http2.producers import h2stream_producer, h2header_producer
from .http2.fifo import priority_producer_fifo
import threading
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO


class http2_request_handler:
	collector = None
	producer = None
	http11_terminator = 24
	def __init__ (self, handler, request):
		self.handler = handler
		self.wasc = handler.wasc
		self.request = request
		self.channel = request.channel
		
		self.channel.producer_fifo = priority_producer_fifo ()
		
		self.conn = H2Connection(client_side = False)
		self.frame_buf = self.conn.incoming_buffer
		self.frame_buf.max_frame_size = self.conn.max_inbound_frame_size
		
		self.data_length = 0
		self.current_frame = None
		self.rfile = BytesIO ()
		self.buf = b""

		self.requests = {}
		self.priorities = {}
		self.promises = {}
		
		self._send_stream_id = 0
		self._closed = False
		self._got_preamble = False

		self._plock = threading.Lock () # for self.conn
		self._clock = threading.Lock () # for self.x
		
	def go_away (self, errcode = 0, msg = None):
		with self._plock:
			self.conn.close_connection (errcode, msg)
		self.send_data ()
		self.channel.close_when_done ()		
				
	def close (self, errcode = 0, msg = None):
		if self._closed: return
		self._closed = True		
		if self.channel:
			self.go_away (errcode) # go_away			
	
	def enter_shutdown_process (self):
		self.close (NO_ERROR)
			
	def closed (self):
		return self._closed
			
	def send_data (self):
		with self._plock:
			data_to_send = self.conn.data_to_send ()
		
		if data_to_send:
			#print ("SEND", repr (data_to_send), len (data_to_send))
			self.channel.push (data_to_send)
	
	def handle_preamble (self):
		if self.request.version.startswith ("2."):
			self.channel.set_terminator (6) # SM\r\n\r\n		
					
	def initiate_connection (self):
		self.handle_preamble ()
		h2settings = self.request.get_header ("HTTP2-Settings")
		if h2settings:
			self.conn.initiate_upgrade_connection (h2settings)
		else:	
			self.conn.initiate_connection()		
		self.send_data ()
		
		if self.request.version == "1.1":
			self.handle_request (1, self.upgrade_header ())			
			
	def upgrade_header (self):
		headers = [
			(":method", self.request.command.upper ()), 
			(":path", self.request.uri),
		]
		for line in self.request.get_header ():
			k, v = line.split (": ", 1) 
			k = k.lower ()
			if k in ("http2-settings", "connection", "upgrade"):
				continue
			headers.append ((k, v))		
		return headers
					
	def collect_incoming_data (self, data):
		#print ("RECV", repr (data), len (data), '::', self.channel.get_terminator ())		
		if not data:
			# closed connection
			self.close (True)
			return
			
		if self.data_length:
			self.rfile.write (data)
		else:
			self.buf += data
	
	def set_frame_data (self, data):
		if not self.current_frame:
			return []
		self.current_frame.parse_body (memoryview (data))				
		self.current_frame = self.frame_buf._update_header_buffer (self.current_frame)
		with self._plock:
			events = self.conn._receive_frame (self.current_frame)
		return events

	def set_terminator (self, terminator):
		self.channel.set_terminator (terminator)
					
	def found_terminator (self):
		buf, self.buf = self.buf, b""
		
		#print ("FOUND", repr (buf), '::', self.data_length, self.channel.get_terminator ())
		#print ("FOUND", self.request.version, self.request.command, self.data_length)
		events = None	
		if self.request.version == "1.1" and self.request.command in ("post", "put") and self.data_length:
			self.request.version = "2.0" # upgrade
			data = self.rfile.getvalue ()
			with self._clock:
				r = self.requests [1]
			r.channel.set_data (data, len (data))

			self.data_length = 0
			self.rfile.seek (0)
			self.rfile.truncate ()			
			self.channel.set_terminator (24) # for premble
		
		elif not self._got_preamble:			
			if not buf.endswith (b"SM\r\n\r\n"):
				raise ProtocolError ("Invalid preamble")
			self._got_preamble = True
			self.channel.set_terminator (9)
		
		elif self.data_length:
			events = self.set_frame_data (self.rfile.getvalue ())
			self.current_frame = None
			self.data_length = 0
			self.rfile.seek (0)
			self.rfile.truncate ()			
			self.channel.set_terminator (9) # for frame header
									
		elif buf:
			self.current_frame, self.data_length = self.frame_buf._parse_frame_header (buf)
			self.frame_buf.max_frame_size = self.conn.max_inbound_frame_size
			self.frame_buf._validate_frame_length (self.data_length)
			
			if self.data_length == 0:
				events = self.set_frame_data (b'')
			self.channel.set_terminator (self.data_length == 0 and 9 or self.data_length)	# next frame header
			
		else:
			raise ProtocolError ("Frame decode error")
		
		if events:
			self.handle_events (events)
	
	def get_new_stream_id (self):
		with self._clock:
			self._send_stream_id += 2
			stream_id = self._send_stream_id
		return stream_id
		
	def push_promise (self, stream_id, request_headers, addtional_request_headers):
		promise_stream_id = self.get_new_stream_id ()
		with self._clock:
			self.promises [promise_stream_id] = request_headers	+ addtional_request_headers	
		
		with self._plock:
			self.conn.push_stream (stream_id, promise_stream_id, request_headers	+ addtional_request_headers)
		self.send_data ()
					
	def push_response (self, stream_id, headers, producer):
		#print ("++RESPONSE", headers)
		with self._clock:
			r = self.requests [stream_id]
			try:
				depends_on, weight = self.priorities [stream_id]
			except KeyError:
				depends_on, weight = 0, 1	
			else:
				del self.priorities [stream_id]				
		
		self.channel.push_with_producer (
			h2header_producer (stream_id, headers, producer, self.conn, self._plock)
		)
		if producer:
			outgoing_producer = h2stream_producer (
				stream_id, depends_on, weight, producer, self.conn, self._plock
			)			
			self.channel.push_with_producer (outgoing_producer)
			
			r.http2 = None # break bacj ref.
			with self._clock:
				#self.channel.ready = self.channel.producer_fifo.ready
				del self.requests [stream_id]			
			
		promise_stream_id, promise_headers = None, None
		with self._clock:
			try: promise_stream_id, promise_headers = self.promises.popitem ()
			except KeyError: pass
		
		if promise_stream_id:
			self.handle_request (promise_stream_id, promise_headers, is_promise = True)
		
	def handle_events (self, events):
		for event in events:
			if isinstance(event, RequestReceived):
				self.handle_request (event.stream_id, event.headers)				
					
			elif isinstance(event, StreamReset):
				if event.remote_reset:
					deleted = False
					if event.stream_id % 2 == 0: # promise stream
						with self._clock:
							try: del self.promises [event.stream_id]
							except KeyError: pass
							else: deleted = True
					
					if not deleted:
						self.channel.producer_fifo.remove (event.stream_id)
						
			elif isinstance(event, ConnectionTerminated):
				self.close (True)
				
			elif isinstance(event, PriorityUpdated):
				if event.exclusive:
					# rebuild depend_ons
					for stream_id in list (self.priorities.keys ()):
						depends_on, weight = self.priorities [stream_id]
						if depends_on == event.depends_on:
							self.priorities [stream_id] = [event.stream_id, weight]
					
				with self._clock:
					self.priorities [event.stream_id] = [event.depends_on, event.weight]
				
			elif isinstance(event, DataReceived):				
				with self._clock:	
					r = self.requests [event.stream_id]
				try:
					r.channel.set_data (event.data, event.flow_controlled_length)
				except ValueError:
					self.close ()
				
			elif isinstance(event, StreamEnded):
				r = None
				with self._clock:
					try: r = self.requests [event.stream_id]
					except KeyError: pass
				
				if r and r.collector:
					r.collector.found_terminator ()
					
		self.send_data ()
							
	def handle_request (self, stream_id, headers, is_promise = False):
		#print ("++REQUEST: %d" % stream_id, headers)
		command = "GET"
		uri = "/"
		scheme = "http"
		authority = ""
		cl = 0
		h = []
		cookies = []
		for k, v in headers:
			#print ('HEADER:', k, v)
			if k[0] == ":":
				if k == ":method": command = v.upper ()
				elif k == ":path": uri = v
				elif k == ":scheme": scheme = v.lower ()
				elif k == ":authority":
					authority = v
					if authority:
						h.append ("host: %s" % authority.lower ())				
				continue								
			if k == "content-length":
				cl = int (v)				
			elif k == "cookie":
				cookies.append (v)
				continue					
			h.append ("%s: %s" % (k, v))
								
		if cookies:
			h.append ("Cookie: %s" % "; ".join (cookies))
		
		should_have_collector = False	
		if command == "CONNECT":
			first_line = "%s %s HTTP/2.0" % (command, authority)
			vchannel = self.channel			
		else:	
			first_line = "%s %s HTTP/2.0" % (command, uri)
			if command in ("POST", "PUT"):
				should_have_collector = True
				vchannel = data_channel (self.channel, cl)
			else:
				self.request.version = "2.0"
				vchannel = fake_channel (self.channel)
		
		r = http2_request (self, vchannel, first_line, command.lower (), uri, "2.0", scheme, h, stream_id, is_promise)		
		vchannel.current_request = r
		
		with self._clock:
			self.channel.request_counter.inc()
			self.channel.server.total_requests.inc()
		
		for h in self.channel.server.handlers:
			if h.match (r):		
				with self._clock:
					self.requests [stream_id] = r
				
				try:					
					h.handle_request (r)
					
				except:
					self.channel.server.trace()
					try: r.response.error (500)
					except: pass
						
				else:
					if should_have_collector and cl > 0 and r.collector is None:
						# too large body
						with self._plock:
							self.conn.reset_stream (stream_id, FLOW_CONTROL_ERROR)
						self.send_data ()	
						# some browser ignore reset, why?
						self.close (FLOW_CONTROL_ERROR)
						
					elif cl > 0:
						if stream_id == 1:
							self.data_length = cl	
							self.set_terminator (self.data_length)
							
						else:
							# give permission for sending data to a client
							with self._plock:
								self.conn.increment_flow_control_window (cl)
								rfcw = self.conn.remote_flow_control_window (stream_id)
								if cl > rfcw:
									self.conn.increment_flow_control_window (cl - rfcw, stream_id)
							self.send_data ()
								
				return					
					
		try: r.response.error (404)
		except: pass


class h2_request_handler (http2_request_handler):
	http11_terminator = None
	
	def handle_preamble (self):
		if self.request.version.startswith ("2."):
			self.conn.receive_data ("PRI * HTTP/2.0\r\n\r\n")
			self.channel.set_terminator (None)
			
	def collect_incoming_data (self, data):		
		with self._plock:
			events = self.conn.receive_data (data)
		self.handle_events (events)


class Handler (wsgi_handler.Handler):
	keep_alive = 120
	
	def match (self, request):
		if request.command == "pri" and request.uri == "*" and request.version == "2.0":
			return True		
		upgrade = request.get_header ("upgrade")
		return upgrade and upgrade.lower () == "h2c" and request.version == "1.1"
	
	def handle_request (self, request):
		http2 = http2_request_handler (self, request)		
		request.channel.die_with (http2, "http2 stream")
		request.channel.set_response_timeout (self.keep_alive)
		request.channel.set_keep_alive (self.keep_alive)
		
		if request.version == "1.1":
			request.response (
				"101 Switching Protocol",
				headers = [("Connection",  "upgrade"), ("Upgrade", "h2c"), ("Server", skitai.NAME.encode ("utf8"))]
			)
			request.response.done (False, False, False, (http2, http2.http11_terminator))
		
		else:
			request.channel.current_request = http2
			
		http2.initiate_connection ()
	


from skitai.saddle import Saddle
import skitai
	
def app (env, start_response):
	start_response ("200 OK", [("Content-Type", "text/plain")])
	return ['Hello World']

skitaiapp = Saddle (__name__)
skitaiapp.use_reloader = True
skitaiapp.debug = True

@skitaiapp.route('/')
def index (was):
	return "Hello World by Saddle"
	
if __name__ == "__main__":
	skitai.run (
		address = "127.0.0.1",
		port = 5000,
		mount = [
			('/', (__file__, 'app')),
			('/skitai', (__file__, 'skitaiapp'))
		],
		#certfile = r"C:\skitaid\etc\cert\skitai.com.ca.pem",
		#passphrase = ""
	)



from . import wsgi_executor
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib


class Executor (wsgi_executor.Executor):
	def __call__ (self):
		request = self.env ["skitai.was"].request
		
		data = self.env ["wsgi.input"].read ()		
		args, methodname = xmlrpclib.loads (data)
		
		if methodname != "system.multicall":
			thunks = [(methodname, args)]
		else:
			thunks = []
			for _methodname, _args in [(each ["methodName"], each ["params"]) for each in args [0]]:
				thunks.append ((_methodname, _args))
		
		self.build_was ()
		
		results = []
		for _method, _args in thunks:
			path_info = self.env ["PATH_INFO"] = "/" + _method.replace (".", "/")			
			current_app, thing, param, respcode = self.find_method (request, path_info)
			
			if respcode:		
				try: raise Exception('Method Error: %s' % _method)
				except: results.append ({'faultCode' : 1, 'faultString' : wsgi_executor.traceback ()})
				continue				
		
			self.was.subapp = current_app
			try:
				result = self.generate_content (thing, _args, {})
			except:
				results.append ({'faultCode' : 1, 'faultString' : self.was.app.debug and wsgi_executor.traceback () or "Error Occured"})
			else:
				results.append ((result [0],))
			del self.was.subapp
		
		if methodname != "system.multicall": results = results [0]
		else: results = (results,)
		
		self.commit ()
		self.was.response ["Content-Type"] = "text/xml"
		
		del self.was.env	
		return xmlrpclib.dumps (results, methodresponse = True, allow_none = True, encoding = "utf8").encode ("utf8")
		



import re
import stat
from . import mime_type_table
import os
from . import filesys
from skitai.server import http_date
from skitai.lib import producers
from skitai.server.utility import *
from hashlib import md5

IF_MODIFIED_SINCE = re.compile (
	'If-Modified-Since: ([^;]+)((; length=([0-9]+)$)|$)',
	re.IGNORECASE
	)

IF_NONE_MATCH = re.compile (
	'If-None-Match: "?(.+?)(?:$|")',
	re.IGNORECASE
	)
	
USER_AGENT = re.compile ('User-Agent: (.*)', re.IGNORECASE)

CONTENT_TYPE = re.compile (
	r"Content-Type: ([^;]+)((; boundary=([A-Za-z0-9'\(\)+_,./:=?-]+)$)|$)",
	re.IGNORECASE
	)

class Handler:
	directory_defaults = [
		'index.htm',
		'index.html'
		]
		
	def __init__ (self, wasc, pathmap = {}, max_age = 300, alt_handlers = []):
		self.wasc = wasc
		self.max_age = max_age
		self.alt_handlers = alt_handlers
		self.permit_cache = {}
		self.filesystem = filesys.mapped_filesystem ()
		for k, v in list(pathmap.items ()):
			self.add_route (k, v)
	
	def close (self):	
		for h in self.alt_handlers:
			try: h.close ()
			except AttributeError: pass
										
	def match (self, request):
		return 1
	
	def add_route (self, alias, option):
		self.filesystem.add_map (alias, option)
	
	def handle_alternative (self, request):
		for h in self.alt_handlers:
			if h.match (request):
				try:
					h.handle_request (request)
				except:						
					self.wasc.logger.trace("server")
					try: request.response.error (500)
					except: pass
				return
		request.response.error (404)
		
	def isprohibited (self, request, path):
		dirname = os.path.split(path) [0]				
		permission = self.filesystem.get_permission (dirname)
		if not permission:
			return False								
		if not self.wasc.authorizer.has_permission (request, permission):
			return True			
		return False	
		
	def handle_request (self, request):
		if request.command not in ('get', 'head'):
			self.handle_alternative (request)
			return
		
		path, params, query, fragment = request.split_uri()
		if '%' in path:
			path = unquote (path)
		
		# return redirect
		if path == "":
			request.response['Location'] = '/'
			request.response.error (301)
			return
		
		if path.find ("./") != -1 or path.find ("//") != -1:
			request.response.error (404)
			return
				
		while path and path [0] == '/':
			path = path[1:]
		
		if self.filesystem.isdir (path):
			if path and path[-1] != '/':
				request.response['Location'] = '/%s/' % path
				request.response.error (301)
				return

			found = False
			if path and path[-1] != '/':
				path = path + '/'
			
			for default in self.directory_defaults:
				p = path + default
				if self.filesystem.isfile (p):
					path = p
					found = True
					break
			
			if not found:
				self.handle_alternative (request)
				return
		
		elif not self.filesystem.isfile (path):
			self.handle_alternative (request)
			return
		
		if self.isprohibited (request, path):
			return
		
		try:
			mtime = self.filesystem.stat (path)[stat.ST_MTIME]
			file_length = self.filesystem.stat (path)[stat.ST_SIZE]
		except:
			self.handle_alternative (request)
			return		
		
		etag = self.make_etag (file_length, mtime)		
		inm = get_header_match (IF_NONE_MATCH, request.header)
		if inm and etag == inm.group (1):
			request.response.start (304)
			request.response.done()
			return
			
		ims = get_header_match (IF_MODIFIED_SINCE, request.header)
		length_match = 1
		if ims:
			length = ims.group (4)
			if length:
				try:
					length = int(length)
					if length != file_length:
						length_match = 0
				except:
					pass
					
		ims_date = 0
		if ims:
			ims_date = http_date.parse_http_date (ims.group (1))

		if length_match and ims_date:
			if mtime <= ims_date:
				request.response.start (304)
				request.response.done()
				return
		
		try:
			file = self.filesystem.open (path, 'rb')
		except IOError:
			self.handle_alternative (request)
			return

		request.response ['Content-Length'] = file_length
		if request.version == "1.0":
			request.response ['Last-Modified'] = http_date.build_http_date (mtime)
			if self.max_age:
				request.response ['Expires'] = http_date.build_http_date (mtime + self.max_age)
		else:
			request.response ['Etag'] = '"' + etag + '"'
			if self.max_age:
				request.response ['Cache-Control'] = "max-age=%d" % self.max_age		
		self.set_content_type (path, request)
		
		if request.command == 'get':
			request.response.push (producers.file_producer (file))			
		request.response.done()
	
	def make_etag (self, file_length, mtime):
		return md5 (("%d:%d" % (file_length, int (mtime))).encode ("utf8")).hexdigest()
	
	def set_content_type (self, path, request):
		ext = get_extension (path).lower()
		if ext in mime_type_table.content_type_map:
			request.response['Content-Type'] = mime_type_table.content_type_map[ext]
		else:
			request.response['Content-Type'] = 'text/plain'
			



from skitai.lib import pathtool
from skitai.server import http_server

class Devnull:
	def write(self, bf): pass
	def close(self): pass
	def flush(self): pass

def update ():
	import os, time
	bladelibpath = os.path.split (os.path.split (pathtool.modpath (http_server)[-1])[0])[0]
	ups = os.path.join (bladelibpath, "blade-pending")
	if not os.path.isdir (ups): return
	today = time.strftime ("%Y%m%d%H%M%S", time.localtime (time.time ()))
	rento = os.path.join (bladelibpath, "blade-bak-%s" % today [:8])
	if os.path.isdir (rento):
		rento = os.path.join (bladelibpath, "blade-bak-%s" % today)		
	os.rename (os.path.join (bladelibpath, "blade"), rento)
	os.rename (os.path.join (bladelibpath, "blade-pending"), os.path.join (bladelibpath, "blade"))
	
def shutdown (root):
	from skitai.server import shutdown
	shutdown.shutdown (root)

def lock (root, redirect_error = 1):
	import os, sys
	from skitai.lib import pathtool
	
	pathtool.mkdir (os.path.join (root, "var/lock"))	
	if redirect_error:
		sys.stderr = open (os.path.join (root, "var/lock/error"), "a")
	
	pidfile = os.path.join (root, "var/lock/pid")
	if os.path.isfile (pidfile):
		raise AssertionError("process is runnig. terminated.")
				
	f = open (pidfile, "w")
	f.write ("%s" % os.getpid ())
	f.close ()

def unlock (root):
	import os
	pidfile = os.path.join (root, "var/lock/pid")
	if os.path.isfile (pidfile):
		os.remove (pidfile)
	
def usage ():
	print((
		"Usage:\n"
		"  %s [--root=path] [--shutdown]\n" % os.path.split (sys.argv [0])[-1]
	))




import threading
import time
from . import asynpsycopg2, asynredis, synsqlite3, asynmongo
from skitai.client import socketpool
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB

class DBPool (socketpool.SocketPool):
		
	def get_name (self):
		return "__dbpool__"
			
	def create_asyncon (self, server, params, dbtype):
		if DB_SQLITE3 in (params [0], dbtype):
			params = ("",) + params [1:]
			return synsqlite3.SynConnect (server, params, self.lock, self.logger)
		
		try: 
			host, port = server.split (":", 1)
		except ValueError:
			host, port = server, 5432
		else:
			port = int (port)
		
		if params [0] == DB_REDIS:
			dbtype = params [0]
			params = ("",) + params [1:]			
		elif params [1] in (DB_MONGODB, DB_PGSQL):	
			dbtype = params [1]
			params = (params [0], "", params [2])
			
		if dbtype == DB_REDIS:
			con_class = asynredis.AsynConnect 
		elif dbtype == DB_MONGODB:			
			con_class = asynmongo.AsynConnect
		elif dbtype == DB_PGSQL:
			con_class = asynpsycopg2.AsynConnect
		return con_class ((host, port), params, self.lock, self.logger)
		
	def get (self, server, dbname, user, pwd, dbtype = DB_PGSQL):
		serverkey = "%s/%s/%s/%s" % (server, dbname, user, dbtype)
		return self._get (serverkey, server, (dbname, user, pwd), dbtype)


if __name__ == "__main__":
	from skitai import lifetime
	from skitai.lib import logger
	from skitai.server.threads import trigger
	
	trigger.start_trigger ()
	pool = DBPool (logger.screen_logger ())
	
	def query ():
		conn = pool.get ("mydb.us-east-1.rds.amazonaws.com:5432", "mydb", "postgres", "")
		conn.execute ("SELECT * FROM cities;")
		rs = conn.fetchwait (5)
		print(rs.status, rs.result)
		
		conn.execute ("INSERT INTO weather VALUES ('San Francisco', 46, 50, 0.25, '1994-11-27');")		
		rs = conn.wait (5)
		print(rs.status, rs.result)
		
		conn.execute ("INSERT INTO weather VALUES ('San Francisco', 54, 67, 0.25, '1994-11-27');")		
		rs = conn.wait (5)
		print(rs.status, rs.result)
		
		
	threading.Thread (target = query).start ()	
	while threading.activeCount () > 1:
		lifetime.loop ()
		


from . import http_response, counter
import time
import re

class http_request:
	version = "1.1"
	collector = None
	producer = None
	request_count = counter.counter()
	
	def __init__ (self, *args):
		self.request_number = self.request_count.inc()		
		(self.channel, self.request,
		 self.command, self.uri, self.version,
		 self.header) = args
		self.logger = self.channel.server.server_logger
		self.server_ident = self.channel.server.SERVER_IDENT		
		self.body = None
		self.reply_code = 200
		self.reply_message = ""		
		self.loadbalance_retry = 0
		self.rbytes = 0
		self.created = time.time ()
		self.gzip_encoded = False
		self.is_promise = False
		self._split_uri = None
		self._header_cache = {}

		self.set_log_info ()
		self.response = http_response.http_response (self)
	
	def set_log_info (self):
		self.gtxid = self.get_header ("X-Gtxn-Id")
		if not self.gtxid:
			self.gtxid = "gtid-%s-c%s-r%s" % (
				self.channel.server.hash_id,
				self.channel.channel_number, 
				self.request_count
			)
			self.ltxid = 1000
		else:			
			self.ltxid = self.get_header ("X-Ltxn-Id")
			if not self.ltxid:
				raise ValueError ("Local txn ID missing")
			self.ltxid = int (self.ltxid) + 1000
			
		self.token = None
		self.claim = None
		self.user = None		
		self.host = self.get_header ("host")
		self.user_agent = self.get_header ("user-agent")
	
	def get_gtxid (self):
		return self.gtxid
		
	def get_ltxid (self, delta = 1):
		self.ltxid += delta
		return str (self.ltxid)
			
	def get_scheme (self):	
		from .https_server import https_channel		
		return isinstance (self.channel, https_channel) and "https" or "http"
	
	def get_raw_header (self):
		return self.header
	get_headers = get_raw_header
	
	path_regex = re.compile (r'([^;?#]*)(;[^?#]*)?(\?[^#]*)?(#.*)?')
	def split_uri (self):
		if self._split_uri is None:
			m = self.path_regex.match (self.uri)
			if m.end() != len(self.uri):
				raise ValueError("Broken URI")
			else:
				self._split_uri = m.groups()				
		return self._split_uri

	def get_header_with_regex (self, head_reg, group):
		for line in self.header:
			m = head_reg.match (line)
			if m.end() == len(line):
				return head_reg.group (group)
		return ''
	
	def set_body (self, body):
		self.body = body
	
	def get_body (self):
		return self.body
	
	def set_header (self, name, value):
		self.header.append ("%s: %s" % (name, value))		
		
	def get_header (self, header = None, default = None):
		if header is None:
			return self.header
		header = header.lower()
		hc = self._header_cache
		if header not in hc:
			h = header + ':'
			hl = len(h)
			for line in self.header:
				if line [:hl].lower() == h:
					r = line [hl:].strip ()
					hc [header] = r
					return r
			hc [header] = None
			return default
		else:
			return hc[header] is not None and hc[header] or default
	
	def get_header_with_attr (self, header, default = None):
		d = {}
		v = self.get_header (header, default)
		if v is None:
			return default, d
			
		v2 = v.split (";")
		if len (v2) == 1:
			return v, d
		for each in v2 [1:]:
			try:
				a, b = each.strip ().split ("=", 1)
			except ValueError:
				a, b = each.strip (), None
			d [a] = b
		return v2 [0], d	
	
	def get_content_length (self):
		try: return int (self.get_header ("content-length"))
		except ValueError: return None
					
	def get_content_type (self):
		return self.get_header_with_attr ("content-type") [0]
				
	def get_main_type (self):
		ct = self.get_content_type ()
		if ct is None:
			return
		return ct.split ("/", 1) [0]
	
	def get_sub_type (self):
		ct = self.get_content_type ()
		if ct is None:
			return
		return ct.split ("/", 1) [1]
		
	def get_user_agent (self):
		return self.get_header ("user-agent")
	
	def get_remote_addr (self):
		return self.channel.addr [0]
			
	def collect_incoming_data (self, data):
		if self.collector:
			self.rbytes += len (data)
			self.collector.collect_incoming_data (data)			
		else:
			self.logger.log (
				'dropping %d bytes of incoming request data' % len(data),
				'warning'
				)

	def found_terminator (self):		
		if self.collector:
			self.collector.found_terminator()			
		else:
			self.logger.log (
				'unexpected end-of-record for incoming request',
				'warning'
				)
			


"""
2008 added for asynchronous DNS query by Hans Roh
"""

import asyncore
import socket
import time
import types
import random
import types
from .pydns import Base, Type, Class, Lib, Opcode
import random

defaults = Base.defaults
Base.DiscoverNameServers ()

class async_dns (asyncore.dispatcher_with_send):
	zombie_timeout = 3
	
	def __init__ (self, servers, request, args, callback, logger, debug_level):
		self.servers = servers
		if not self.servers:
			raise OSError ("DNS Server Required")
		random.shuffle (self.servers)
		#self.servers = [("156.154.71.3", 513), ("156.154.71.9", 543)]
		self.addr = self.servers.pop (0)
		self.request = request
		self.callback = callback		
		self.args = args		
		self.logger = logger
		self.debug_level = debug_level
		self.qname = self.args ["name"].decode ("utf8")
		
		self.creation_time = time.time ()
		self.event_time = time.time ()
		self.ac_in_buffer = b""
		self.closed = False
		
		asyncore.dispatcher_with_send.__init__ (self)		
		self.create_socket (socket.AF_INET, socket.SOCK_STREAM)				
		self.ac_out_buffer = self.request
				
		try:
			self.connect (self.addr)
		except:
			self.handle_error ()
	
	def __repr__ (self):
		return "<async_dns: %s>" % self.args ["name"]
				
	def trace (self):
		self.logger.trace ()
	
	def log_info (self, line, level = 'info'):
		self.log ("[%s] %s" % (level, line))
	
	def log (self, line):
		self.logger (line)	
	
	def create_socket (self, family, type):
		if hasattr (socket, "_no_timeoutsocket"):
			sock_class = socket._no_timeoutsocket
		else:
			sock_class = socket.socket
			
		self.family_and_type = family, type
		self.socket = sock_class (family, type)
		self.socket.setblocking (0)
		self._fileno = self.socket.fileno ()
		self.add_channel()
		
	def handle_error (self):
		if self.debug_level: 
			self.trace ()
		self.close ()
	
	def handle_timeout (self):
		self.handle_close ()
					
	def handle_connect (self):	
		self.send (self.request)
		
	def handle_write (self):
		self.socket.shutdown (1)
		
	def handle_read (self):
		data = self.recv (4096)
		self.ac_in_buffer += data		
		
	def handle_expt (self):
		self.handle_close ()
	
	def close (self):
		if self.closed:
			return
		self.closed = True	
		asyncore.dispatcher_with_send.close (self)
		self.callback (self.servers, self.request, self.args, self.ac_in_buffer)
			
	def handle_close (self):
		self.close ()
	
	
class Request:
	def __init__(self, logger, server = [], debug_level = 1):
		if type (server) is bytes:
			server = [server]
		
		defaults ["server"] += server
		self.logger = logger
		self.debug_level = debug_level		
		
	def argparse (self, name, args):
		args['name']=name
		for i in list(defaults.keys()):
			if i not in args:
				if i == "server": 
					args[i]=defaults[i][:]
				else:	
					args[i]=defaults[i]
				
		if type (args['server']) == bytes:
			args ['server'] = [args['server']]
			
		return args
		
	def req (self, name, **args):
		name = name.encode ("utf8")
		args = self.argparse (name, args)
		
		protocol = args ['protocol']
		port = args ['port']
		opcode = args ['opcode']
		rd = args ['rd']
		server = args ['server'][:]
		
		if type(args['qtype']) in (bytes, str):
			try:
				qtype = getattr (Type, args ['qtype'].upper ())
			except AttributeError:
				raise Base.DNSError('%s unknown query type' % name)
				
		else:
			qtype = args ['qtype']
			
		qname = args ['name']
		
		#print 'QTYPE %d(%s)' % (qtype, Type.typestr(qtype))
		m = Lib.Mpacker()
		# jesus. keywords and default args would be good. TODO.
		m.addHeader(0,
			  0, opcode, 0, 0, rd, 0, 0, 0,
			  1, 0, 0, 0)
		
		m.addQuestion (qname, qtype, Class.IN)
		request = m.getbuf ()
		request = Lib.pack16bit (len(request)) + request
		
		server = [(x, args ["port"]) for x in server]
		async_dns (server, request, args, self.processReply, self.logger, self.debug_level)
			
	def processReply(self, server, request, args, data):		
		exception = 0
		
		try:
			if not data:
				raise Base.DNSError('%s, no working nameservers found' % args ['name'])
		
			if args ["protocol"] == "tcp":
				header = data [:2]
				if len (header) < 2:
					raise Base.DNSError('%s, EOF' % args ['name'])
				count = Lib.unpack16bit(header)		
				reply = data [2: 2 + count]					
				if len (reply) != count:
					raise Base.DNSError('%s, incomplete reply' % args ['name'])						
			
			else:
				reply = data
			
		except:
			if server:				
				async_dns (server, request, args, self.processReply, self.logger, self.debug_level)
				return
				
			else:
				reply = None
		
		if reply is None:		
			answers = []
			
		else:
			try:	
				u = Lib.Munpacker(reply)
				r = Lib.DnsResult(u, args)
				r.args = args
				answers = r.answers
				
			except:
				if self.debug_level: self.logger.trace ()
				answers = []	

		callback = args.get ("callback", None)
		if callback:
			if type (callback) != type ([]):
				callback = [callback]
			
			for cb in callback:
				cb (answers)
			

if __name__	== "__main__":
	from skitai.lib import logger
	import pprint
	f = Request (logger.screen_logger ())
	f.req ("www.openfos.com", protocol = "tcp", callback = pprint.pprint, qtype="a")
	f.req ("www.yahoo.com", protocol = "tcp", callback = pprint.pprint, qtype="a")
	f.req ("www.hungryboarder.com", protocol = "tcp", callback = pprint.pprint, qtype="a")
	f.req ("www.alexa.com", protocol = "tcp", callback = pprint.pprint, qtype="a")
	f.req ("www.google.com", protocol = "tcp", callback = pprint.pprint, qtype="mx")
	asyncore.loop (timeout = 1)
	
	



"""
 $Id: Type.py,v 1.6 2002/03/19 12:41:33 anthonybaxter Exp $

 This file is part of the pydns project.
 Homepage: http://pydns.sourceforge.net

 This code is covered by the standard Python License.

 TYPE values (section 3.2.2)
"""

A = 1           # a host address
NS = 2          # an authoritative name server
MD = 3          # a mail destination (Obsolete - use MX)
MF = 4          # a mail forwarder (Obsolete - use MX)
CNAME = 5       # the canonical name for an alias
SOA = 6         # marks the start of a zone of authority
MB = 7          # a mailbox domain name (EXPERIMENTAL)
MG = 8          # a mail group member (EXPERIMENTAL)
MR = 9          # a mail rename domain name (EXPERIMENTAL)
NULL = 10       # a null RR (EXPERIMENTAL)
WKS = 11        # a well known service description
PTR = 12        # a domain name pointer
HINFO = 13      # host information
MINFO = 14      # mailbox or mail list information
MX = 15         # mail exchange
TXT = 16        # text strings
AAAA = 28       # IPv6 AAAA records (RFC 1886)
SRV = 33        # DNS RR for specifying the location of services (RFC 2782)


# Additional TYPE values from host.c source

UNAME = 110
MP = 240

# QTYPE values (section 3.2.3)

AXFR = 252      # A request for a transfer of an entire zone
MAILB = 253     # A request for mailbox-related records (MB, MG or MR)
MAILA = 254     # A request for mail agent RRs (Obsolete - see MX)
ANY = 255       # A request for all records

# Construct reverse mapping dictionary

_names = dir()
typemap = {}
for _name in _names:
    if _name[0] != '_': typemap[eval(_name)] = _name

def typestr(type):
    if type in typemap: return typemap[type]
    else: return repr(type)
#
# $Log: Type.py,v $
# Revision 1.6  2002/03/19 12:41:33  anthonybaxter
# tabnannied and reindented everything. 4 space indent, no tabs.
# yay.
#
# Revision 1.5  2002/03/19 12:26:13  anthonybaxter
# death to leading tabs.
#
# Revision 1.4  2001/08/09 09:08:55  anthonybaxter
# added identifying header to top of each file
#
# Revision 1.3  2001/07/19 07:38:28  anthony
# added type code for SRV. From Michael Strer.
#
# Revision 1.2  2001/07/19 06:57:07  anthony
# cvs keywords added
#
#



# Web Socket message dump and load related codes are based on websocket-server 0.4
# by Johan Hanssen Seferidis
# https://pypi.python.org/pypi/websocket-server
#
# 2016. 1. 16 Modified by Hans Roh

from . import wsgi_handler
from hashlib import sha1
from base64 import b64encode
from skitai.server.http_response import catch
from skitai.server import utility
from skitai import version_info, was as the_was
import threading		
from .websocket import specs
from .websocket.servers import websocket_servers
import time

class Handler (wsgi_handler.Handler):
	def match (self, request):
		upgrade = request.get_header ("upgrade")
		return upgrade and upgrade.lower ().startswith ("websocket") and request.version == "1.1" and request.command == "get"
	
	def close (self):
		websocket_servers.close ()
	
	GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'.encode ()
	def calculate_response_key (self, key):		
		hash = sha1(key.encode() + self.GUID)
		response_key = b64encode(hash.digest()).strip()
		return response_key.decode('ASCII')
		
	def handle_request (self, request):				
		def donot_response (self, *args, **kargs):
			def push (thing):
				raise AssertionError ("Websocket can't use start_response ()")
			return push	
		
		origin = request.get_header ("origin")
		host = request.get_header ("host")
		protocol = request.get_header ("http_sec_websocket_protocol", 'unknown')
		securekey = request.get_header ("sec-webSocket-key")
		
		if not origin or not host or not securekey: 
			return request.response.error (400)
		
		if not (host.startswith ("localhost:") or origin.find (host) != -1 or origin == "null"):
			return request.response.error (403)
		
		path, params, query, fragment = request.split_uri ()		
		has_route = self.apps.has_route (path)
		if type (has_route) is int:
			return request.response.error (404)
		
		apph = self.apps.get_app (path)
		if not self.isauthorized (apph.get_callable(), request):
			return
			
		env = self.build_environ (request, apph)
		was = the_was._get ()
		was.request = request
		env ["skitai.was"] = was
		env ["websocket_init"] = ""
		# app should reply  (design type one of (1,2,3), keep-alive seconds)
		# when env has 'skitai.websocket_init'
		try:
			apph (env, donot_response)
			try:
				design_spec, keep_alive, param_name = env ["websocket_init"]
				del env ["websocket_init"]
			except (IndexError, ValueError): 
				raise AssertionError ("You should return (design_spec, keep_alive, param_name) where env has key 'skitai.websocket_init'")				
			assert design_spec in (1,2,3,4), "design_spec  should be one of (WEBSOCKET_REQDATA, WEBSOCKET_DEDICATE, WEBSOCKET_DEDICATE_THREADSAFE, WEBSOCKET_MULTICAST)"			
		except:
			self.wasc.logger.trace ("server",  request.uri)
			return request.response.error (500, why = apph.debug and catch (1) or "")
		
		headers = [
			("Sec-WebSocket-Accept", self.calculate_response_key (securekey)),
			("Upgrade", "Websocket"),
			("Connection", "Upgrade"),
      ("WebSocket-Protocol", protocol),
      ("WebSocket-Location", "ws://" + host + path)
		]
		request.response ("101 Web Socket Protocol Handshake", headers = headers)
		
		if design_spec == 1: 
			# WEBSOCKET_REQDATA			
			# Like AJAX, simple request of client, simple response data
			# the simplest version of stateless HTTP protocol using basic skitai thread pool
			ws = specs.WebSocket1 (self, request, apph, env, param_name)
			env ["websocket"] = ws		
			self.channel_config (request, ws, keep_alive)
		
		elif design_spec in (2, 4): 
			# WEBSOCKET_DEDICATE 			
			# 1:1 wesocket:thread
			# Be careful, it will be consume massive thread resources
			if design_spec == 2:
				ws = specs.WebSocket2 (self, request)
			else:
				ws = specs.WebSocket4 (self, request)
			request.channel.die_with (ws, "websocket spec. %d" % design_spec)
			env ["websocket"] = ws
			self.channel_config (request, ws, keep_alive)
			job = specs.Job2 (request, apph, (env, donot_response), self.wasc.logger)
			threading.Thread (target = job).start ()
		
		else: 
			# WEBSOCKET_MULTICAST
			# /chat?roomid=456, 
			# return (WEBSOCKET_MULTICAST, 600, "roomid")
			# websocketserver thread will be created by roomid
			# can send to all clients of group / specific client
			if not param_name:
				gidkey = path
			else:	
				gid = utility.crack_query (query).get (param_name, None)
				try:
					assert gid, "%s value can't find" % param_name
				except:
					self.wasc.logger.trace ("server",  request.uri)
					return request.response.error (500, why = apph.debug and catch (1) or "")
				gid = "%s/%s" % (path, gid)
			
			if not websocket_servers.has_key (gid):
				server = websocket_servers.create (gid)
				request.channel.die_with (server, "websocket spec. %d" % design_spec)
				env ["websocket"] = server
				job = specs.Job3 (server, request, apph, (env, donot_response), self.wasc.logger)
				threading.Thread (target = job).start ()	
			
			server = websocket_servers.get (gid)				
			ws = specs.WebSocket3 (self, request, server)
			server.add_client (ws)
			self.channel_config (request, ws, keep_alive)			
		
	def channel_config (self, request, ws, keep_alive):
		request.response.done (False, False, False, (ws, 2))
		request.channel.set_response_timeout (keep_alive)
		request.channel.set_keep_alive (keep_alive)
	
	


import threading
import sched, time, re
import bisect, sys, os
from skitai.lib  import confparse, logger as logger_module, processutil
import subprocess
import signal
import win32process, win32api, win32con, pywintypes
try:
	from urllib.request import urlopen
	from urllib.parse import urlparse
except ImportError:
	from urllib import urlopen	
	from urlparse import urlparse
	
import tempfile
				
class Scheduler (sched.scheduler):
	def __init__(self, wasc, config_path = None, logger = None, cache = 100):
		self.wasc = wasc
		self.logger = logger
		self.queue = []
		self.actions = {}
		self.jobs = {}
		self.timefunc = time.time
		self.delayfunc = time.sleep		
		self.active = 1
		self.supressed = 0
		self.supresstime = time.time ()
		self.config_path = config_path
		if not os.path.isfile (self.config_path):
			f = open (self.config_path, "w")
			f.write ("")
			f.close ()
		self.cv = threading.Condition (threading.RLock())
		self.mutate = 0
		self.run = 0
		self.log_path = tempfile.mkdtemp ()
		
		for each in self.logger.loggers:
			if isinstance (each, logger_module.rotate_logger):
				self.log_path = each.base
		
		if config_path:
			self.enter_from_config (config_path)
			
		f = threading.Thread (target = self.loop)
		f.setDaemon (1)
		f.start ()
	
	def makeBatchFile (self, batch, command):
		if os.name == "nt":
			shell_script_ext = ".cmd"	
		else:
			shell_script_ext = ".sh"
			
		instance_home = self.wasc.instance_home
		blade_home = self.wasc.blade_home
		blade_module = os.path.split (self.wasc.blade_home) [0]
		
		path = os.path.join (instance_home, "bin/%s%s" % (batch, shell_script_ext))
		env = (
			"REM @set PYTHONPATH=%%PYTHONPATH%%;%s\r\n"
			"REM @set BLADE_HOME=%s\r\n"
			"REM @set BLADE_INSTANCE_HOME=%s\r\n" % (blade_module, blade_home, instance_home)
		)		
		f = open (path, "w")
		f.write (env + command + " %1 %2 %3 %4 %5 %6 %7 %8 %9\r\n")		
		f.close ()
		
		return path
		
	def supress (self, sec = 30):
		self.supresstime = time.time () + sec
		self.supressed = 1
	
	def unsupress (self):
		self.supressed = 0	
		
	def mktime (self, date):
		date += [0, 0, 0]
		return time.mktime (tuple (date))
			
	def calculate (self, dates):		
		runtimes = []
		for date in dates.split (','):		
			week = -1
			current = self.timefunc()
			parsecurrent = time.localtime (current)			
			if date.find ("/") > -1:
				date, week = date.split ("/")
				try: week = int (week)
				except: week = -1
				if not (0 <= week < 7): 
					week = -1
			date = date.strip ()	
			parsedate = re.split ('\s+', date)
			
			if len (parsedate) < 3: continue
			for i in range (6 - len (parsedate)):
				parsedate.append ('0')
			
			wishdate = []
			for i in range (6):
				if parsedate [i] == '*':
					wishdate.append (parsecurrent [i])
				else:
					wishdate.append (int (parsedate [i]))
					
			nextrun = self.mktime (wishdate [:])
			
			# date control
			got_date = 0
			if nextrun > current:
				got_date = 1
			
			else:				
				for i in range (4, -1, -1):
					twishdate = wishdate [:]					
					if parsedate [i] == '*':
						twishdate [i] = twishdate [i] + 1
						nextrun = self.mktime (twishdate)
						if nextrun > current:
							wishdate = twishdate [:6]
							got_date = 1
							break
							
			if not got_date: return
			
			# date control
			date = time.localtime (self.mktime (wishdate [:3] + [0] * 3)) [:3]
			#apply recalculate date
			wishdate = list (date) + wishdate [3:]
			today = parsecurrent [:3]
			
			if date > today:
				for i in range (3):
					if date [i] != today [i]: break
						
				for j in range (4, i, -1):
					if parsedate [j] == '*':
						if j in (1, 2): wishdate [j] = 1
						else: wishdate [j] = 0
						
				nextrun = self.mktime (wishdate [:6])
			
			if week != -1:
				cweek = int (time.strftime ("%w", time.localtime (nextrun)))
				while week != cweek:
					nextrun += 3600 * 24
					cweek = int (time.strftime ("%w", time.localtime (nextrun)))
				
			runtimes.append (nextrun)
							
		if runtimes:
			runtimes.sort ()
			return runtimes [0]
	
	def enterabs (self, *event):
		bisect.insort(self.queue, event)
		return event
	
	def remove (self, job_id):
		self.cv.acquire ()
		self.mutate = 1
		queue = []
		for event in self.queue:
			if job_id == event [1]: continue
			bisect.insort (queue, event)
		self.queue = queue
		self.mutate = 0
		self.cv.notifyAll ()
		self.cv.release ()
	
	def enable (self, job_id, enable = 1, runafter = 0):		
		if job_id not in self.actions: return
		self.actions [job_id][-3] = enable
		self.renter (job_id, runafter)
			
	def renter (self, job_id, runafter = 0):
		if job_id not in self.actions: return		
		job_name, args, interval, date, enable, startat, timeouttokil = self.actions [job_id]
		self.enter (job_id, job_name, args, interval, date, enable, runafter, startat, timeouttokil)
		
	def enter (self, job_id, job_name, args, interval, date, enable = 1, runafter = 0, startat = "", timeouttokill = ""):
		try: timeouttokill = float (timeouttokill)
		except: timeouttokill = 0.	
		self.actions [job_id] = [job_name, args, interval, date, enable, startat, timeouttokill]
		self.remove (job_id)
		
		if runafter > 0:
			self.enterabs (self.timefunc() + (runafter * 60), job_id)
			return			
		
		nextrun = None		
		if date:
			nextrun = self.calculate (date)
			if interval:
				self.actions [job_id][3] = None # first starting time and run with interval
		elif interval:
			nextrun  = self.timefunc() + (interval * 60)
			
		if nextrun:
			self.enterabs (nextrun, job_id)
		else:
			self.actions [job_id] = self.actions [job_id][-3] = 0 # disable
			self.enterabs (sys.maxsize, job_id)
	
	RX_ENV = re.compile ("(%([_a-zA-Z]+)%)")
	def replace_env (self, text):
		for env, key in self.RX_ENV.findall (text):
			if os.environ.get (key, ""):
				text = text.replace (env, os.environ [key])
		return text
	
	def url_open (self, job_id, url, timeout):
		# Don't self https requesting!
		# I don't know why, but server will be hung up
		job_name = self.actions [job_id][0]
		self.logger ("[info] schedule %s started" % job_name)
		try:
			scheme, netloc, script, params, qs, fragment = urlparse (url)
			call = self.wasc.rcall.Server ("%s://%s" % (scheme, netloc))
			uri = script
			if params: uri += ";" + params
			if qs: uri += "?" + qs
			call.request (uri)
			self.jobs [job_id] = [None, None, 0, time.time ()]
			rs = call.getwait (timeout)
			self.logger ("[info] schedule %s result status: %s, http response code: %s" % (job_name, rs.status, rs.status == 1 and rs.code or -1))
		
		except:
			self.logger.trace (job_name)
		
		try:				
			del self.jobs [job_id]
		except KeyError:	
			pass
	
	def url_open2 (self, job_id, url, timeout):
		# Due to timeoputsocket, https will not work
		job_name = self.actions [job_id][0]
		self.logger ("[info] schedule %s started" % job_name)
		try:
			self.jobs [job_id] = [None, None, 0, time.time ()]
			rs = urlopen (url, timeout = timeout)			
			self.logger ("[info] schedule %s result status: %s %s" % (job_name, rs.code, rs.msg))		
		except:
			self.logger.trace (job_name)
		
		try:				
			del self.jobs [job_id]
		except KeyError:	
			pass
		
	def create_process (self, job_id, cmd, startat):
		job_name = self.actions [job_id][0]
		try:	
			dlogs = []
			for file in os.listdir (self.log_path):
				if file.find ("schedtask-%s-" % job_id) != 0: continue
				dlogs.append (file)	
			
			if len (dlogs) > 7:
				dlogs.sort ()
				for file in dlogs [:-7]:
					try: os.remove (os.path.join (self.log_path, file))
					except: pass
		
			output = os.path.join (self.log_path, "task-%s.log" % job_id)
			if os.path.isfile (output):
				rotate = open (os.path.join (self.log_path, "task-%s-%s.log" % (job_id, logger_module.now (0))), "a")
				olds = open (output)
				for line in olds:
					rotate.write (line)					
				olds.close ()
				rotate.close ()
			
			fh = open (output, "w")
			
			
				
			cmd = self.replace_env (cmd)
			if startat:
				cmdline = 'cmd.exe /C "cd /d %s && %s"' % (startat, cmd)
			else:	
				cmdline = 'cmd.exe /C "%s"' % cmd
				
			proc = subprocess.Popen (cmdline, stdout=fh, stderr=fh)
			
		except: 
			self.logger.trace ()
		
		else:		
			self.logger ("[info] schedule %s started" % job_name)
			self.jobs [job_id] = [proc, fh, 0, time.time ()]
	
	def isrunning (self, job_id):		
		if job_id not in self.jobs: return 0
		proc, fh, sizef, lastmodified = self.jobs [job_id]
		if proc:
			status = proc.poll ()
			if status is None: 
				return 1
			else:
				del self.jobs [job_id]
				if fh:
					fh.close()
				job_name = self.actions [job_id][0]
				self.logger ("[info] schedule %s terminated" % job_name)	
				return 0
		else:
			return 1 # maybe url request
	
	def kill (self, job_id):	
		if job_id not in self.jobs: 
			return 0, "", "no such job %s" % job_id
		job_name = self.actions [job_id][0]
		proc, fh, sizef, lastmodified = self.jobs [job_id]
		if proc is None: # url request
			return 0, "", "job %s is not a process" % job_name
			
		cpids = [proc.pid] + processutil.get_child_pid (proc.pid)
		cpids.reverse ()
		for cpid  in cpids:
			try:
				handle = win32api.OpenProcess (win32con.PROCESS_TERMINATE, 0, cpid)
				win32api.TerminateProcess (handle, 0)
				win32api.CloseHandle (handle)		
			except pywintypes.error as why:
				if why.errno == 87:
					pass
				else:	
					return why, "", "killing %s(pid:%d) has been failed" % (job_name, proc.pid)
		
		del self.jobs [job_id]
		if fh: fh.close()
		return 0, "", "%s(pid:%d) was killed" % (job_name, proc.pid)
	
	def maintern (self):	
		for job_id in list(self.jobs.keys ()):
			job_name = self.actions [job_id][0]
			lastmodified = self.jobs [job_id][-1]
			timeouttokill = self.actions [job_id][-1]
			if timeouttokill and time.time () - lastmodified > (timeouttokill * 60) and self.isrunning (job_id):
				self.logger ("[info] schedule %s was timeout, try to kill..., result: %s" % (job_name, self.kill (job_id)))
				
	def loop (self):
		while 1:
			self.run += 1
			if self.run % 10 == 0:
				self.maintern ()
			
			if not self.active or not self.queue:
				self.delayfunc(3)
				continue
			
			if self.supressed:
				if self.supresstime > time.time ():
					self.delayfunc(3)					
					continue
				else:
					self.supressed = 0	
			
			self.cv.acquire ()
			runnable = False
			while self.mutate: self.cv.wait ()			
			for i in range (len(self.queue)):
				sched, job_id = self.queue [i]
				if not self.isrunning (job_id):
					runnable = True
					break
			self.cv.release ()
			if not runnable:
				continue
					
			remain = sched - self.timefunc()			
			if remain > 0.:
				if remain <= 3.:
					self.delayfunc (remain)
				else:
					self.delayfunc (3)
				continue
			
			job_name, args, interval, date, enable, startat, timeouttokill = self.actions [job_id]			
			if enable:
				if args.startswith ("HTTP "):
					threading.Thread (target = self.url_open, args = (job_id, args[5:].strip (), timeouttokill)).start ()
				elif args.startswith ("SHELL "):
					self.create_process (job_id, args[6:].strip (), startat)
				else:
					self.create_process (job_id, args, startat)

			self.renter (job_id)
	
	def readoutput (self, job_id, char = 10000):
		fo = os.path.join (self.log_path, "task-%s.log" % job_id)
		if not os.path.isfile (fo): return []
		
		size = os.stat (fo).st_size		
		if job_id in self.jobs:		
			self.jobs [job_id][-2] = size
			self.jobs [job_id][-1] = time.time ()
					
		if size < char: pos = 0
		else: pos = size - char
		f = open (fo)
		f.seek (pos)		
		if pos > 0: f.readline ()		
		buffer = []
		while 1:
			t = f.readline ()
			if not t: return buffer
			buffer.append (t.strip ())
		f.close ()	
		return buffer

	def cancel (self):
		self.active = 0
	
	def start (self):
		self.active = 1
	
	def status (self):
		if not self.active:
			asctime = time.asctime (time.localtime (time.time ()))			
			return [('-1', 'All jobs stopped', '', 'Scheduler')]
		
		k = []
		for sched, job_id in self.queue:
			if job_id not in self.actions: continue
			job_name, args, interval, date, enable, startat, timeouttokill = self.actions [job_id]
			if sched == sys.maxsize:
				asctime = "Not Scheduled"
			else:
				asctime = time.asctime (time.localtime (sched))				
			
			if self.isrunning (job_id) and not enable:
				status = "be disabled"
			elif self.isrunning (job_id):
				status = "running"
			elif enable:
				status = "waiting"				
			else:
				status = "disabled"
				
			k.append ((job_id, job_name, asctime, status, args))
					
		return k
	
	def add_job (self, job_id):		
		job_name = self.config.getopt (job_id, 'name')
		args = self.config.getopt (job_id, 'args')		
		date = self.config.getopt (job_id, 'date')
		interval = self.config.getint (job_id, 'interval')
		enable = self.config.getint (job_id, 'enable')
		runafter = self.config.getint (job_id, 'runatstart')
		timeouttokill = self.config.getopt (job_id, 'timeouttokill')
		if timeouttokill is None: timeouttokill = ""
		startat = self.config.getopt (job_id, 'startat')
		if startat is None: startat = ""		
		self.enter (job_id, job_name, args, interval, date, enable, runafter, startat, timeouttokill)
		
	def enter_from_config (self, config_path):
		self.config = confparse.ConfParse (config_path)
		for job_id in list(self.config.keys ()):
			self.add_job (job_id)
		
	def remove_config (self, job_id):
		self.remove (job_id)
		
		if job_id in self.config:
			del self.config [job_id]
			self.config.update ()			
		
		try: del self.actions [job_id]	
		except KeyError: pass		
		
	valid_keys = ['name','args','date','interval','enable','runatstart','timeouttokill','startat']
	def new_config (self, job_id, dict):
		self.config.makesect (job_id)
		for k, v in list(dict.items ()):
			if k not in self.valid_keys: continue
			self.config.setopt (job_id, k, v)
		self.config.update ()
		self.add_job (job_id)
	
	def has_key (self, job_id):
		return job_id in self.config
	
	def get_config (self):
		return self.config
			
	def update_config (self, job_id, dict):
		for k, v in list(dict.items ()):
			if k not in self.valid_keys: continue
			self.config.setopt (job_id, k, v)
		self.config.update ()
		self.add_job (job_id)
		
	def cleanup (self):	
		self.active = 0		
		


import sys, os
from skitai.lib import pathtool
import importlib

try: _reloader = importlib.reload
except AttributeError: _reloader = reload
	
def importer (directory, libpath):
	sys.path.insert(0, directory)	
	__import__ (libpath, globals ())
	libpath, abspath = pathtool.modpath (libpath)
	module = sys.modules [libpath]	
	if abspath [-4:] in (".pyc", ".pyo"):
		abspath = abspath [:-1]		
	sys.path.pop (0)		
	return module, abspath

def reimporter (module):
	directory, libpath = os.path.split (module.__file__)[0], module.__name__
	del sys.modules [module.__name__]
	importer (directory, libpath)
	
def reloader (module):
	directory = os.path.split (module.__file__) [0]
	sys.path.insert(0, directory)
	_reloader (module)
	sys.path.pop (0)		
	
	



import time
from skitai.lib import jwt

class AuthorizedUser:
	def __init__ (self, user, roles, realm):
		self.name = user	
		self.roles = roles
		self.realm = realm
		
class Handler:
	def __init__ (self, wasc, authenticate, realm, proxypass_handler, secret_key = None):
		self.wasc = wasc
		self.authenticate = authenticate
		self.realm = realm
		self.proxypass_handler = proxypass_handler
		self.secret_key = secret_key
		if type (self.secret_key) is str:
			self.secret_key = self.secret_key.encode ("utf8")
		self.auth_handler = None
		
	def set_auth_handler (self, storage):
		self.auth_handler = storage
	
	def set_error (self, request, error = "", desciption = ""):		
		if error:
			request.response ["WWW-Authenticate"] = 'Bearer realm="%s", error="%s", error_description="%s"' % (self.realm, error, desciption)
		else:
			request.response ["WWW-Authenticate"] = 'Bearer realm="%s"' % self.realm
		
	def match (self, request):
		if self.proxypass_handler.match (request):
			return 1
		return 0
			
	def handle_request (self, request):
		if not self.authenticate:
			return self.continue_request (request)
		
		authorization = request.get_header ("authorization")
		if not authorization or authorization[:7].lower () != "bearer ":
			self.set_error (request)
			return request.response.abort (401)
		
		token = authorization [7:]
		if token.find (".") != -1: # JWT
			if not self.secret_key:
				self.set_error (request, "secret_key_error", "Secret key error")
				return request.response.abort (500)
			claim = jwt.get_claim (self.secret_key, token)
			if not claim:
				self.set_error (request, "invalid_token", "The access token invalid")
				return request.response.abort (401)			
			request.token = token.split (".")[1]
			if self.auth_handler and hasattr (self.auth_handler, "handle_claim"):					
				request.claim = claim
				self.auth_handler.handle_claim (request, self.continue_request)					
			else:
				self.continue_request (request, claim.get ("user"), claim.get ("roles"))
			
		else:
			if self.auth_handler is None or not hasattr (self.auth_handler, "handle_token"):
				self.set_error (request, "token_handler_error", "Token handler error")
				return request.response.abort (500)			
			request.token = token
			self.auth_handler.handle_token (request, self.continue_request)
	
	def continue_request (self, request, username = None, roles = None):
		if self.authenticate:			
			request.user = AuthorizedUser (username, roles, self.realm)			
			if not roles:
				self.set_error (request, "invalid_token", "The access token not exists")
				return request.response.abort (401)
			
			cluster = self.proxypass_handler.find_cluster (request)[0]
			if not cluster.has_permission (request, roles):
				self.set_error (request, "insufficient_scope", "The access token is not valid")
				return request.response.abort (403)
		
		self.proxypass_handler.handle_request (request)
		
		


"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""

import skitai
__VER__ = skitai.VERSION

import sys
import os
import shutil, glob
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

if sys.argv[-1] == 'publish':
	if os.name == "nt":
		os.system('python setup.py sdist upload') # bdist_wininst --target-version=2.7
	else:
		os.system('python setup.py sdist upload')
	sys.exit()

classifiers = [
  'License :: OSI Approved :: BSD License',
  'Development Status :: 4 - Beta',
  'Topic :: Internet :: WWW/HTTP',
	'Topic :: Internet :: WWW/HTTP :: HTTP Servers',				
	'Environment :: Console',
	'Environment :: No Input/Output (Daemon)',
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'Programming Language :: Python',
	'Programming Language :: Python :: 2.7',
	'Programming Language :: Python :: 3'
]

PY_MAJOR_VERSION = sys.version_info [0]
if PY_MAJOR_VERSION == 3:
	if os.path.isfile ("skitai/lib/py2utils.py"):
		os.remove ("skitai/lib/py2utils.py")
		try: os.remove ("skitai/lib/py2utils.pyc")
		except OSError: pass		
else:
	if not os.path.isfile ("skitai/lib/py2utils.py"):
		with open ("skitai/lib/py2utils.py", "w") as f:
			f.write ("def reraise(type, value, tb):\n\traise type, value, tb\n")			

packages = [
	'skitai',
	'skitai.client',
	'skitai.dbapi',
	'skitai.lib',	
	'skitai.protocol',
	'skitai.examples',
	'skitai.protocol.dns',
	'skitai.protocol.dns.pydns',
	'skitai.protocol.http',	
	'skitai.protocol.http2',
	'skitai.protocol.ws',	
	'skitai.protocol.smtp',
	'skitai.server',
	'skitai.server.dbi',
	'skitai.server.handlers',
	'skitai.server.rpc',
	'skitai.server.threads',
	'skitai.server.handlers.http2',
	'skitai.server.handlers.websocket',
	'skitai.server.handlers.proxy',
	'skitai.saddle'
]

package_dir = {
	'skitai': 'skitai',
	'skitai.server': 'skitai/server'
}

skitai_files = [
	"protocol/dns/*.txt",
	"protocol/dns/pydns/*.txt"
]

package_data = {
	"skitai": skitai_files
}

setup(
	name='skitai',
	version=__VER__,
	description='Skitai Library',
	url = 'https://gitlab.com/hansroh/skitai',
	author='Hans Roh',
	author_email='hansroh@gmail.com',	
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='BSD',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/skitai",
	install_requires = ["jinja2==2.8", "h2", "redis>=2.10", "pymongo"],
	classifiers=classifiers
)



from jinja2 import environment
from jinja2 import lexer as olexer
from jinja2 import parser
from jinja2 import nodes
import re


# patch for enabling line_statement_prefix raw
from jinja2.lexer import TOKEN_DATA, TOKEN_RAW_END, Failure, \
	TOKEN_WHITESPACE, TOKEN_FLOAT, TOKEN_INTEGER, TOKEN_NAME, TOKEN_STRING, TOKEN_OPERATOR, \
	whitespace_re, float_re, integer_re, name_re, string_re, operator_re, compile_rules


def get_lexer (environment):
	c = lambda x: re.compile(x, re.M | re.S)
	e = re.escape
	tag_rules = [
	    (whitespace_re, TOKEN_WHITESPACE, None),
	    (float_re, TOKEN_FLOAT, None),
	    (integer_re, TOKEN_INTEGER, None),
	    (name_re, TOKEN_NAME, None),
	    (string_re, TOKEN_STRING, None),
	    (operator_re, TOKEN_OPERATOR, None)
	]	
	root_tag_rules = compile_rules(environment)
	block_suffix_re = environment.trim_blocks and '\\n?' or ''
	block_prefix_re = '%s' % e(environment.block_start_string)
	prefix_re = {}
	
	if environment.lstrip_blocks:
		no_lstrip_re = e('+')
		block_diff = c(r'^%s(.*)' % e(environment.block_start_string))
		m = block_diff.match(environment.comment_start_string)
		no_lstrip_re += m and r'|%s' % e(m.group(1)) or ''
		m = block_diff.match(environment.variable_start_string)
		no_lstrip_re += m and r'|%s' % e(m.group(1)) or ''
		
		comment_diff = c(r'^%s(.*)' % e(environment.comment_start_string))
		m = comment_diff.match(environment.variable_start_string)
		no_variable_re = m and r'(?!%s)' % e(m.group(1)) or ''
		
		lstrip_re = r'^[ \t]*'
		block_prefix_re = r'%s%s(?!%s)|%s\+?' % (
		        lstrip_re,
		        e(environment.block_start_string),
		        no_lstrip_re,
		        e(environment.block_start_string),
		        )
		comment_prefix_re = r'%s%s%s|%s\+?' % (
		        lstrip_re,
		        e(environment.comment_start_string),
		        no_variable_re,
		        e(environment.comment_start_string),
		        )
		prefix_re['block'] = block_prefix_re
		prefix_re['comment'] = comment_prefix_re
		
	else:
		block_prefix_re = '%s' % e(environment.block_start_string)

	lexer = olexer.get_lexer(environment)
	lexer.rules['root'] = [
			# directives
			(c('(.*?)(?:%s)' % '|'.join(
				[r'(?P<raw_begin>(?:\s*%s\-|%s|^[ \t\v]*%s)\s*raw\s*(?:\-%s\s*|%s|\s*:?\s*))' % (
					e(environment.block_start_string),
					block_prefix_re,
					e(environment.line_statement_prefix),
					e(environment.block_end_string),
					e(environment.block_end_string)
				)] + [
					r'(?P<%s_begin>\s*%s\-|%s)' % (n, r, prefix_re.get(n,r))
					for n, r in root_tag_rules
				])), (TOKEN_DATA, '#bygroup'), '#bygroup'),
			# data
			(c('.+'), TOKEN_DATA, None)
		]
	
	lexer.rules [olexer.TOKEN_RAW_BEGIN] = [
			(c('(.*?)((?:\s*%s\-|%s|^[ \t\v]*%s)\s*endraw\s*(?:\-%s\s*|%s%s|$))' % (
				e(environment.block_start_string),
				block_prefix_re,
				e(environment.line_statement_prefix),
				e(environment.block_end_string),
				e(environment.block_end_string),
				block_suffix_re
			)), (TOKEN_DATA, TOKEN_RAW_END), '#pop'),
			(c('(.)'), (Failure('Missing end of raw directive'),), None)
		]		
	return lexer
	
	
class Parser (parser.Parser):			
	def subparse(self, end_tokens=None):
		body = []
		data_buffer = []
		add_data = data_buffer.append

		if end_tokens is not None:
			self._end_token_stack.append(end_tokens)

		def flush_data():
			if data_buffer:
				lineno = data_buffer[0].lineno
				body.append(nodes.Output(data_buffer[:], lineno=lineno))
				del data_buffer[:]

		try:
			while self.stream:
				token = self.stream.current
				if token.type == 'data':
					if token.value:
						add_data(nodes.TemplateData(token.value,
													lineno=token.lineno))
					next(self.stream)
				elif token.type == 'variable_begin':
					next(self.stream)
					# added by Hans Roh 2016.5.1
					# If variable_begin & variable_end is #,
					# should excape by #"#"#
					# this patch make it easy to ##
					if self.stream.current.type == "variable_end":
						add_data(nodes.TemplateData(token.value, lineno=token.lineno))
						next(self.stream)
					else:	
						add_data(self.parse_tuple(with_condexpr=True))
						self.stream.expect('variable_end')
				elif token.type == 'block_begin':
					flush_data()
					next(self.stream)
					if end_tokens is not None and \
					   self.stream.current.test_any(*end_tokens):
						return body
					rv = self.parse_statement()
					if isinstance(rv, list):
						body.extend(rv)
					else:
						body.append(rv)
					self.stream.expect('block_end')
				else:
					raise AssertionError('internal parsing error')

			flush_data()
		finally:
			if end_tokens is not None:
				self._end_token_stack.pop()

		return body


class Environment (environment.Environment):
	def _parse(self, source, name, filename):
		return Parser(self, source, name, environment.encode_filename(filename)).parse()
	lexer = property(get_lexer, doc="The lexer for this environment.")

# enable 'raw' line statement
environment.Environment = property(get_lexer, doc="The lexer for this environment.")



from . import wsgi_executor
try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
import sys, os
import struct
from skitai.lib import producers
from collections import Iterable

    
class grpc_producer (producers.iter_producer):		
	def more (self):		
		try:
			data = next (self.data)
			serialized = data.SerializeToString ()
			return struct.pack ("<B", 0) + struct.pack ("<i", len (serialized)) + serialized			
		except StopIteration:			
			return b""
	
	
class Executor (wsgi_executor.Executor):
	def get_messages (self, fp):
		msgs = []
		byte = fp.read (1)
		while byte:
			iscompressed = struct.unpack ("<B", byte) [0]
			length = struct.unpack ("<i", fp.read (4)) [0]
			msg = fp.read (length)
			msgs.append (msg)
			byte = fp.read (1)
		return tuple (msgs)
		
	def __call__ (self):
		request = self.env ["skitai.was"].request
		msgs = self.get_messages (self.env.get ("wsgi.input"))
		servicename, methodname = self.env ["PATH_INFO"].split ("/") [-2:]
		
		self.build_was ()
		current_app, thing, param, respcode = self.find_method (request, "/" + methodname)
		if respcode: 
			return b""
		
		self.was.subapp = current_app
		result = self.generate_content (thing, msgs, {})		
		del self.was.subapp
		
		self.commit ()
		self.was.response ["Content-Type"] = "application/grpc+proto"

		del self.was.env		
		
		if type (result) is list:
			result = iter (result)		
		return grpc_producer (result)
		
		


from skitai.server import http_request
from .response import response
import time

class request (http_request.http_request):
	def __init__ (self, *args):
		self.request_number = self.request_count.inc()		
		(self.http2, self.channel, self.request,		 
		 self.command, self.uri, self.version, self.scheme,
		 self.header, self.stream_id, self.is_promise) = args
		
		self.logger = self.channel.server.server_logger
		self.server_ident = self.channel.server.SERVER_IDENT
		self.body = None
		self.reply_code = 200
		self.reply_message = ""		
		self._split_uri = None
		self._header_cache = {}
		self.rbytes = 0
		self.created = time.time ()
		self.loadbalance_retry = 0
		self.gzip_encoded = False		
		
		self.outgoing_producer = None
		self.depends_on = 0
		self.weight = 0
		
		self.set_log_info ()				
		self.response = response (self)
	
	def get_scheme (self):	
		return self.scheme
		

	
	



from . import wsgi_handler, proxy_handler
from skitai.protocol.http import request as http_request
import re


class Handler (proxy_handler.Handler):
	def __init__ (self, wasc, clusters, cachefs = None):
		proxy_handler.Handler.__init__ (self, wasc, clusters, cachefs)
		self.route_map = {}
		self.sorted_route_map = []
	
	def add_route (self, route, cname):		
		self.route_map [route] = (cname, len (route))
		temp = list (self.route_map.items ())
		temp.sort (key = lambda x: x [1][1], reverse = True)
		self.sorted_route_map = temp
		
	def match (self, request):		
		return self.find_cluster (request) and 1 or 0
				
	def find_cluster (self, request):
		uri = request.uri
		for route, (cname, route_len) in self.sorted_route_map:
			if uri.startswith (route):
				return self.clusters [cname], route_len
		
	def will_open_tunneling (self):
		return self.response.code == 101 # websocket connection upgrade

	def continue_request (self, request, collector):
		request.set_header ("X-GTXN-ID", request.gtxid)
		request.set_header ("X-LTXN-ID", request.ltxid)
		request.loadbalance_retry = 0
		if self.has_valid_cache (request, collector is not None):
			return
		
		try:
			self.route (request, collector)			
		except:
			self.wasc.logger.trace ("server")	
			request.response.error (500, "", "Routing failed. Please contact administator.")
	
	def route (self, request, collector):
		current_cluster, route_len  = self.find_cluster (request)
		psysicaluri = request.uri [route_len:]
		if psysicaluri == "": psysicaluri = "/"
		elif psysicaluri[0] != "/": psysicaluri = "/" + psysicaluri
		
		request.loadbalance_retry += 1
		retry = request.loadbalance_retry
		asyncon = current_cluster.get (index = -1)
		
		if not asyncon: 
			request.logger ("nopool-%d, url: %s, cluster index: %d" % (
				retry, 
				request.uri, 
				route
				), 'warn')
			return request.response.error (503)
		
		if retry > 1:
			request.logger ("failsafe-%d, target: %s:%s, url: %s, cluster index: %d" % (
				retry,
				asyncon.address [0], 
				asyncon.address [1],				
				request.uri,
				route
			), 'warn')
		
		fetcher = http_request.HTTPRequest (psysicaluri, request.command, collector is not None, logger = self.wasc.logger.get ("server"))		
		r = proxy_handler.proxy_request_handler (asyncon, fetcher, self.callback, request, collector)	
		r.handle_request ()
	
	def callback (self, handler):
		request, response, collector = handler.client_request, handler.response, handler.collector		
		cluster = self.find_cluster (request) [0]
		cluster.report (handler.asyncon, response.code)	
		
		if response.code >= 700:
			if request.loadbalance_retry >= len (cluster):
				request.response.error (506, response.msg)
				
			elif request.channel:
				if not collector or (collector and collector.cached):
					if collector:
						collector.reuse_cache ()
					self.route (request, collector)
					return self.dealloc (request, handler)
		
		else:
			try:	
				self.save_cache (request, handler)					
			except:
				self.wasc.logger.trace ("server")
		
		request.loadbalance_retry = 0		
		self.dealloc (request, handler)		
		